from castle.cms import defaults
from castle.cms.tiles.base import BaseImagesTile
from castle.cms.tiles.base import DisplayTypeTileMixin
from castle.cms.tiles.content import IImagesTileSchema
from castle.cms.tiles.views import BaseTileView
from castle.cms.tiles.views import TileViewsSource
from castle.cms.widgets import PreviewSelectFieldWidget
from plone.autoform import directives as form
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
import zope.schema as schema
from zope.schema.vocabulary import SimpleVocabulary


class DefaultView(BaseTileView):
    name = 'default'
    order = 0
    index = ViewPageTemplateFile('templates/gallery/default.pt')
    tile_name = 'gallery'


class GalleryTile(BaseImagesTile, DisplayTypeTileMixin):
    display_type_name = 'gallery'
    display_type_default = 'default'
    display_type_fallback_view = DefaultView

    render = DisplayTypeTileMixin.render_display

    @property
    def gallery_height(self):
        return (self.data or {}).get('gallery_height', 'default')


class IGalleryTileSchema(IImagesTileSchema):

    form.widget('display_type', PreviewSelectFieldWidget,
                tile_name='gallery')
    display_type = schema.Choice(
        title=u"Display Type",
        source=TileViewsSource('gallery'),
        default=defaults.get('gallery_tile_displaytype', u'default')
    )

    gallery_height = schema.Choice(
        title=u"Gallery Height",
        default='default',
        vocabulary=SimpleVocabulary([
            SimpleVocabulary.createTerm('small', 'small', 'Small'),
            SimpleVocabulary.createTerm('default', 'default', 'Default'),
            SimpleVocabulary.createTerm('large', 'large', 'Large'),
            SimpleVocabulary.createTerm('x-large', 'x-large', 'X-Large')
        ])
    )

    show_thumbnails = schema.Bool(
        title=u'Show Thumbnails',
        description=(
            u'Display thumbnail images below main gallery image. '
            u'This option is for advanced display types in certain add-ons.Usually, this option has no affect.'
        ),
        missing_value=True,
        required=False,
    )

    show_image_title = schema.Bool(
        title=u'Show Image Title',
        description=(
            u'Display Image Title directly below the main gallery image. '
            u'This option is for advanced display types in certain add-ons. Usually, this option has no affect.'
        ),
        missing_value=True,
        required=False,
    )

    show_image_captions = schema.Bool(
        title=u'Show Image Captions',
        description=(
            u'Display a small caption (the image\'s summary field) below either the image title (if visible) or the main image. '
            u'This option is for advanced display types in certain add-ons. Usually, this option has no affect.'
        ),
        missing_value=True,
        required=False,
    )

    show_image_counter =  schema.Bool(
        title=u'Show Image Counter',
        description=(
            u'Display a small indicator above the main image indicating which image is being viewed (ex: "Image 15 of 22"). '
            u'This option is for advanced display types in certain add-ons. Usually, this option has no affect.'
        ),
        missing_value=False,
        required=False,
    )

    show_link_to_original_image =  schema.Bool(
        title=u'Show Link to Original Image',
        description=(
            u'When the gallery image is highlighted, show a link to the original image. '
            u'This option is for advanced display types in certain add-ons. Usually, this option has no affect.'
        ),
        missing_value=False,
        required=False,
    )

    # use_custom_image_captions = schema.Bool(
    #     title=u'Use Custom Image Captions',
    #     description=(
    #         u'The default behavior when Image Captions are enabled is to use the summary '
    #         u'contained on each image item as the caption. This option allows a custom '
    #         u'caption to be specified for each image above.'
    #     ),
    #     missing_value=False,
    #     required=False,
    # )




from plone.api import portal
from Products.CMFPlone.resources.browser.cook import cookWhenChangingSettings
from Products.CMFCore.utils import getToolByName

import importlib


PROFILE_ID = 'profile-castle.cms:default'


def cook_js_resources(context, logger=None):
    cookWhenChangingSettings(portal.get())


def custom_upgrade_factory(profile_id):
    upgrade_file = importlib.import_module('castle.cms.upgrades.upgrade_' + profile_id)
    return upgrade_file.upgrade


def default_upgrade_factory(profile_id):
    def upgrade_function(site, logger=None):
        setup = getToolByName(site, 'portal_setup')
        full_profile_id = 'profile-castle.cms:{}'.format(profile_id)
        setup.runAllImportStepsFromProfile(full_profile_id)
        cookWhenChangingSettings(portal.get())
    return upgrade_function


def profileless_upgrade_factory():
    def upgrade(context, logger=None):
        cookWhenChangingSettings(portal.get())
    return upgrade


upgrade_3000 = default_upgrade_factory('3000')
upgrade_3001 = default_upgrade_factory('3001')
upgrade_3003 = default_upgrade_factory('3003')
upgrade_3004 = default_upgrade_factory('3004')



from lxml import etree
from plone.app.uuid.utils import uuidToObject
from Products.Five import BrowserView
from urlparse import urlparse, parse_qs


class ParallaxView(BrowserView):

    def get_tiles(self):
        parallax_tiles = []
        root = etree.fromstring(self.context.content)
        tiles = root.findall('.//{http://www.w3.org/1999/xhtml}div[@data-tile]')
        for tile in tiles:
            tile_path = tile.get('data-tile')
            if 'castle.cms.parallaxtile/' in tile_path:
                parsed = urlparse(tile_path)
                data = parse_qs(parsed.query)
                parallax_tile = {}

                # Parallax tile properties
                try:
                    image = uuidToObject(data['image:list'][0])
                    parallax_tile['image'] = image.absolute_url()
                except Exception:
                    parallax_tile['image'] = None

                parallax_tile['bg_color'] = data.get('bg_color', ['white'])[0]
                parallax_tile['title'] = data.get('title', [None])[0]
                parallax_tile['title_size'] = data.get('title_size', ['36px'])[0]
                parallax_tile['text'] = data.get('text:text', [None])[0]
                parallax_tile['text_size'] = data.get('text_size', ['12px'])[0]
                hor_position = data.get('hor_text_position', ['center'])[0]
                if hor_position == 'start':
                    parallax_tile['hor_left'] = '5%'
                    parallax_tile['hor_right'] = '45%'
                elif hor_position == 'end':
                    parallax_tile['hor_left'] = '45%'
                    parallax_tile['hor_right'] = '5%'
                else:
                    parallax_tile['hor_left'] = '10%'
                    parallax_tile['hor_right'] = '10%'
                    parallax_tile['justify_center'] = True
                parallax_tile['text_color'] = data.get('text_color', ['black'])[0]
                parallax_tile['text_shadow'] = data.get('text_shadow', [None])[0]

                # Static tile properties
                try:
                    image = uuidToObject(data['static_image:list'][0])
                    parallax_tile['static_image'] = image.absolute_url()
                except Exception:
                    parallax_tile['static_image'] = None

                parallax_tile['static_bg_color'] = data.get('static_bg_color', ['white'])[0]
                parallax_tile['static_title'] = data.get('static_title', [None])[0]
                parallax_tile['static_title_size'] = data.get('static_title_size', ['36px'])[0]
                parallax_tile['static_text'] = data.get('static_text:text', [None])[0]
                parallax_tile['static_text_size'] = data.get('static_text_size', ['12px'])[0]
                hor_position = data.get('static_hor_text_position', ['center'])[0]
                if hor_position == 'start':
                    parallax_tile['static_hor_left'] = '5%'
                    parallax_tile['static_hor_right'] = '45%'
                elif hor_position == 'end':
                    parallax_tile['static_hor_left'] = '45%'
                    parallax_tile['static_hor_right'] = '5%'
                else:
                    parallax_tile['static_hor_left'] = '10%'
                    parallax_tile['static_hor_right'] = '10%'
                    parallax_tile['static_justify_center'] = True
                parallax_tile['static_text_color'] = data.get('static_text_color', ['black'])[0]
                parallax_tile['static_text_shadow'] = data.get('static_text_shadow', [None])[0]

                parallax_tiles.append(parallax_tile)

        return parallax_tiles



from plone.app.textfield import RichText
from plone.autoform.interfaces import IFormFieldProvider
from plone.dexterity.interfaces import IDexterityContent
from plone.supermodel import model
from zope.component import adapter
from zope.interface import implementer
from zope.interface import provider


@provider(IFormFieldProvider)
class ICustomMarkup(model.Schema):
    custom_markup = RichText(
        title=u'Custom Markup',
        description=u'HTML Markup to be included in some templates and listings',
        default_mime_type='text/html',
        output_mime_type='text/html',
        allowed_mime_types=('text/html',),
        required=False,
    )


@implementer(ICustomMarkup)
@adapter(IDexterityContent)
class CustomMarkup(object):
    def __init__(self, context):
        self.context = context



from plone.dexterity.browser import edit
from Products.Five import BrowserView
from lxml import etree
from urlparse import urlparse, parse_qs
from plone.app.uuid.utils import uuidToObject
from plone.api.portal import get as get_portal
from plone.api.portal import get_registry_record
from plone.api.portal import set_registry_record
from Products.CMFPlone.resources import add_resource_on_request


class SlideshowView(BrowserView):

    def get_slides(self):
        slides = []
        root = etree.fromstring(self.context.content)
        tiles = root.findall('.//{http://www.w3.org/1999/xhtml}div[@data-tile]')
        for tile in tiles:
            tile_path = tile.get('data-tile')
            if 'castle.cms.slide/' in tile_path:
                parsed = urlparse(tile_path)
                data = parse_qs(parsed.query)
                slide = {}
                try:
                    image = uuidToObject(data['image:list'][0])
                    slide['image'] = image.absolute_url()
                except Exception:
                    slide['image'] = None
                try:
                    video = uuidToObject(data['video:list'][0])
                    slide['video'] = video.absolute_url()
                except Exception:
                    slide['video'] = None
                slide['displayType'] = data.get('display_type', ['background-image'])[0]
                if slide['displayType'] == 'resource-slide':
                    try:
                        related_uuids = data['related_items:list']
                        slide['related_resources'] = [uuidToObject(uuid) for uuid in related_uuids]
                    except Exception:
                        pass
                slide['title'] = data.get('title', [None])[0]
                slide['text'] = data.get('text:text', [None])[0]
                slide['vert'] = data.get('vert_text_position', ['middle'])[0]
                hor_text_position = data.get('hor_text_position', ['center'])[0]
                slide['hor'] = hor_text_position
                slide['width_class'] = 'full-width' if hor_text_position == 'center' else 'half-width'
                justify_wrapped_text = data.get('justify_wrapped_text:boolean', [0])[0] == '1'
                slide['justify_wrapped_text'] = justify_wrapped_text
                slide['justify_wrap_class'] = 'justify-wrap' if justify_wrapped_text else ''
                slide['text_alignment'] = data.get('text_alignment', ['center'])[0]
                slide['customize_left_slide_mobile'] = \
                    data.get('customize_left_slide_mobile:boolean', [0])[0] == '1'
                if slide['customize_left_slide_mobile']:
                    slide['left_mobile_hor'] = \
                        data.get('left_slide_mobile_hor_text_position', ['default'])[0]
                    slide['left_mobile_vert'] = \
                        data.get('left_slide_mobile_vert_text_position', ['default'])[0]
                    slide['left_mobile_alignment'] = \
                        data.get('left_slide_mobile_text_alignment', ['default'])[0]
                slides.append(slide)
        return slides

    def get_id(self):
        return self.context.custom_dom_id or None

    def get_view_more_link_text(self):
        try:
            return self.context.view_more_link_text
        except Exception:
            return None

    def get_view_more_link_url(self):
        try:
            return self.context.view_more_link_url
        except Exception:
            return None

    def display_view_more_link(self):
        try:
            show_view_more = self.context.show_view_more_link
        except Exception:
            show_view_more = True
        return (show_view_more and
                self.get_view_more_link_text() and
                self.get_view_more_link_url())

    def get_textlines(self, slide):
        slide_text = slide.get('text', '')
        if slide_text:
            return slide_text.split('\n')
        else:
            return None

    def get_resource_view_url(self, resource):
        view_types = get_registry_record('plone.types_use_view_action_in_listings') or []
        url = resource.absolute_url()
        if resource.portal_type in view_types:
            url += '/view'
        return url

    def get_domain(self):
        full_url = get_portal().absolute_url()
        domain = full_url.replace('https://', '').replace('http://', '')
        return domain

    def get_left_slide_mobile_options(self, slide):
        if slide.get('customize_left_slide_mobile', False):
            hor = slide.get('left_mobile_hor', 'default')
            vert = slide.get('left_mobile_vert', 'default')
            align = slide.get('left_mobile_alignment', 'default')
        else:
            hor = 'default'
            vert = 'default'
            align = 'default'
        return{
            'mobile_hor': slide.get('hor', 'center') if hor == 'default' else hor,
            'mobile_vert': slide.get('vert', 'middle') if vert == 'default' else vert,
            'mobile_alignment': slide.get('text_alignment', 'center') if align == 'default' else align,
        }


class SlideshowEditForm(edit.DefaultEditForm):
    def __call__(self, *args, **kw):
        add_resource_on_request(self.request, 'castle-components-slide')
        return super(SlideshowEditForm, self).__call__(*args, **kw)

    def update(self):
        super(SlideshowEditForm, self).update()

    def applyChanges(self, data):
        self.handle_update_default(data, 'link_text')
        self.handle_update_default(data, 'link_url')
        super(SlideshowEditForm, self).applyChanges(data)

    def handle_update_default(self, data, type):
        should_update_registy = data.get('update_default_{}'.format(type), False)
        form_value = data.get('view_more_{}'.format(type), None)
        if should_update_registy and form_value:
            record_name = 'castle.resource_slide_view_more_{}'.format(type)
            try:
                default_value = get_registry_record(record_name)
            except Exception:
                return
            if form_value != default_value:
                set_registry_record(record_name, form_value)
        data['update_default_{}'.format(type)] = False


SlideshowEditView = SlideshowEditForm



from Products.Five import BrowserView
from plone import api


class ExclusionPanel(BrowserView):
    def get_excluded_content(self):
        catalog = api.portal.get_tool('portal_catalog')
        excluded_content = catalog(exclude_from_search=True)
        return excluded_content


class PublishedWithPrivateParents(BrowserView):
    def get_published_private_parent_content(self):
        catalog = api.portal.get_tool('portal_catalog')
        content = catalog(review_state='published', has_private_parents=True)
        return content



"""
Env Vars used:

CASTLE_CMS_AUDIT_LOG_CONFIG_FILE  # specifies a path to a json/dict formatted config for logging
CASTLE_CMS_AUDIT_LOG_INSTANCE     # specifies 'instance' value in audit log schema


Full schema for an audit log message:

schema_version: str  # schema version of log message
schema_type: str     # schema type ("castle.cms.audit")
instance: str        # determined by env var CASTLE_CMS_AUDIT_LOG_INSTANCE
site: str            # either '(zope root)' or the root site path (e.g. '/Castle')
type: str            # audit log type (e.g. 'content', 'workflow', etc)
actionname: str      # previously specified as just 'name', specifies which action for the type
summary: str         # optional short string describing the action
user: str            # username of the account that performed the action
request_uri: str     # the URI of the request that initiated the action
date: str            # ISO date/time the action was performed
object: str          # UUID of object that was affected by action
path: str            # path of object that was affected by action
es2id: Optional[str] # original ES2.x assigned _id -- only present in records migrated from ES2.x storage

"""
from datetime import datetime
import json
import logging
import logging.config
import os


from plone import api
from plone.app.iterate.interfaces import IAfterCheckinEvent
from plone.app.iterate.interfaces import ICancelCheckoutEvent
from plone.app.iterate.interfaces import ICheckoutEvent
from plone.app.iterate.interfaces import IWorkingCopyDeletedEvent
from plone.registry.interfaces import IRecordAddedEvent
from plone.registry.interfaces import IRecordRemovedEvent
from plone.registry.interfaces import IRecordModifiedEvent
from plone.uuid.interfaces import IUUID
from Products.DCWorkflow.interfaces import IAfterTransitionEvent
from Products.PluggableAuthService.interfaces.events import ICredentialsUpdatedEvent
from Products.PluggableAuthService.interfaces.events import IPrincipalCreatedEvent
from Products.PluggableAuthService.interfaces.events import IPrincipalDeletedEvent
from Products.PluggableAuthService.interfaces.events import IPropertiesUpdatedEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedOutEvent
import transaction
from zope.globalrequest import getRequest
from zope.interface import providedBy
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectCopiedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectMovedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent


from castle.cms.events import ICacheInvalidatedEvent
from castle.cms.events import IMetaTileEditedEvent
from castle.cms.events import ITrashEmptiedEvent
from castle.cms.interfaces import ITrashed


logger = logging.getLogger("Plone")

DEFAULT_AUDIT_LOGGER_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'auditlog': {
            'format': '%(asctime)s %(levelname)s %(name)s %(es2id)s %(schema_version)s %(schema_type)s "%(instance)s" "%(site)s" %(type)s "%(actionname)s" "%(summary)s" %(user)s %(request_uri)s %(date)s %(object)s %(path)s',  # noqa
        }
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'auditlog',
        },
    },
    'loggers': {
        'auditlogger': {
            'handlers': ['console'],
            'propagate': False,
            'level': 'INFO',
        },
    },
}

# if this path exists, and loads fine as json, then apply the json as logging
# config (or try to) -- otherwise just pump to stdout
logging_config_file = os.getenv("CASTLE_CMS_AUDIT_LOG_CONFIG_FILE", None)
if logging_config_file is not None and os.path.exists(logging_config_file):
    try:
        with open(logging_config_file, "r") as fin:
            configdict = json.load(fin)
    except Exception:
        configdict = DEFAULT_AUDIT_LOGGER_CONFIG
        logger.error(
            "Couldn't load configuration for auditlogger python logger, "
            "defaulting to stdout", exc_info=True)
else:
    configdict = DEFAULT_AUDIT_LOGGER_CONFIG

try:
    # note: with 'disable_existing_loggers' set to False, this shouldn't wipeout
    # the config from Plone, etc
    logging.config.dictConfig(configdict)
except Exception:
    logger.error("failed to configure audit logger", exc_info=True)
auditlogger = logging.getLogger("auditlogger")


def get_index_name():
    return os.getenv("CASTLE_CMS_AUDIT_LOG_INDEXNAME", "gelfs-*")


class DefaultRecorder(object):
    valid = True

    def __init__(self, data, event, obj=None):
        self.data = data
        self.event = event
        self.obj = obj

    def __call__(self):
        try:
            user = api.user.get_current().getId()
        except api.exc.CannotGetPortalError:
            # likely means that we are logged in on the zope root, not through a site
            from AccessControl import getSecurityManager
            userobj = getSecurityManager().getUser()
            user = userobj.getUserName()

        try:
            requri = getRequest().URL
        except AttributeError:
            # this might occur on, e.g., tests that don't set up a fake request well
            requri = "(none)"

        data = {
            'es2id': None,  # this is from es2.x conversion. new logs should not have this.
            'type': self.data._type,
            'actionname': self.data.name,
            'summary': self.data.summary,
            'user': user,
            'date': datetime.utcnow().isoformat(),
            'request_uri': requri
        }
        if self.obj is not None:
            data['object'] = IUUID(self.obj)
            data['path'] = '/'.join(self.obj.getPhysicalPath())
        else:
            data['object'] = ''
            data['path'] = ''
        return data


class LoggedInRecorder(DefaultRecorder):
    def __call__(self):
        data = super(LoggedInRecorder, self).__call__()
        data['user'] = self.event.object.getId()
        return data


class WorkflowRecorder(DefaultRecorder):

    @property
    def valid(self):
        return self.event.transition is not None

    def __call__(self):
        data = super(WorkflowRecorder, self).__call__()
        data['summary'] = 'previous: %s, new: %s, transition: %s' % (
            self.event.old_state.title,
            self.event.new_state.title,
            self.event.transition.title,
        )
        return data


class MetaTileRecorder(DefaultRecorder):
    def __call__(self):
        data = super(MetaTileRecorder, self).__call__()
        data['summary'] = 'Slot edited: {}'.format(self.event.tile_id)
        return data


class ConfigModifyRecorder(DefaultRecorder):
    def __call__(self):
        data = super(ConfigModifyRecorder, self).__call__()
        try:
            data['summary'] = 'Configuration Record '
            '%s modified. Old value: %s, New value: %s' % (
                self.event.record,
                self.event.oldValue,
                self.event.newValue
            )
        except AttributeError:
            data['summary'] = 'Configuration Record '
            '%s modified.' % self.event.record
        return data


class CacheInvalidatedRecorder(DefaultRecorder):
    def __call__(self):
        data = super(CacheInvalidatedRecorder, self).__call__()
        if self.event.object.success:
            data['summary'] = 'The following urls have been purged: '
            '%s' % self.event.object.purged
        else:
            data['summary'] = 'Cache invalidation failure. '
            'Make sure caching proxies are properly configured.'
        return data


class ContentTypeChangeNoteRecorder(DefaultRecorder):
    def __call__(self):
        data = super(ContentTypeChangeNoteRecorder, self).__call__()
        data['summary'] = 'Change Note Summary: %s' % \
                          self.event.object.changeNote
        return data


class AuditData(object):
    def __init__(self, _type, name, summary=None, recorder_class=DefaultRecorder):
        self._type = _type
        self.name = name
        self.summary = summary
        self.recorder_class = recorder_class

    def get_recorder(self, event, obj=None):
        return self.recorder_class(self, event, obj)


_registered = {
    IAfterCheckinEvent: AuditData('working copy support', 'Check in'),
    ICheckoutEvent: AuditData('working copy support', 'Check out'),
    ICancelCheckoutEvent: AuditData('working copy support', 'Cancel checkout'),
    IWorkingCopyDeletedEvent: AuditData(
        'working copy support', 'Working copy deleted'),
    IObjectAddedEvent: AuditData('content', 'Created'),
    IObjectCopiedEvent: AuditData('content', 'Copied'),
    IObjectModifiedEvent: AuditData(
        'content', 'Modified',
        recorder_class=ContentTypeChangeNoteRecorder),
    IObjectMovedEvent: AuditData('content', 'Moved'),
    IObjectRemovedEvent: AuditData('content', 'Deleted'),
    IPrincipalCreatedEvent: AuditData('user', 'Created'),
    ITrashed: AuditData('content', 'Trashed'),
    IUserLoggedInEvent: AuditData(
        'user', 'Logged in',
        recorder_class=LoggedInRecorder),
    IUserLoggedOutEvent: AuditData('user', 'Logged out'),
    IPrincipalDeletedEvent: AuditData('user', 'Deleted'),
    ICredentialsUpdatedEvent: AuditData('user', 'Password updated'),
    IPropertiesUpdatedEvent: AuditData('user', 'Properties updated'),
    IAfterTransitionEvent: AuditData(
        'workflow', 'Transition',
        recorder_class=WorkflowRecorder),
    IMetaTileEditedEvent: AuditData(
        'slots', 'Slot edited',
        recorder_class=MetaTileRecorder),
    IRecordAddedEvent: AuditData(
        'configuration', 'Added',
        recorder_class=ConfigModifyRecorder),
    IRecordModifiedEvent: AuditData(
        'configuration', 'Modified',
        recorder_class=ConfigModifyRecorder),
    IRecordRemovedEvent: AuditData(
        'configuration', 'Removed',
        recorder_class=ConfigModifyRecorder),
    ITrashEmptiedEvent: AuditData('content', 'Trash Emptied'),
    ICacheInvalidatedEvent: AuditData(
        'content', 'Cache Invalidated',
        recorder_class=CacheInvalidatedRecorder)
}


def record(success, recorder, site_path):
    if not success:
        return
    if recorder.valid:
        data = recorder()
        data["schema_version"] = "1"
        data["schema_type"] = "castle.cms.audit"
        data["instance"] = os.getenv("CASTLE_CMS_AUDIT_LOG_INSTANCE", "(not configured)")
        data["site"] = site_path
        auditlogger.info(site_path, extra=data)


def event(obj, event=None):
    if event is None:
        # some events don't include object
        event = obj
        obj = None

    iface = providedBy(event).declared[0]
    if iface not in _registered:
        return

    if obj is not None and ITrashed.providedBy(obj):
        # special handling for ITrashed...
        # we don't want to record transition events for trashing
        # and we want a special psuedo trash event
        # then, we still want an event when it gets deleted for good
        if iface not in (IAfterTransitionEvent, IObjectRemovedEvent):
            # dive out if not IAfterTransitionEvent or object removed event
            return
        if iface == IObjectRemovedEvent:
            audit_data = _registered[iface]
        else:
            audit_data = _registered[ITrashed]
    else:
        audit_data = _registered[iface]

    recorder = audit_data.get_recorder(event, obj)
    try:
        site_path = '/'.join(api.portal.get().getPhysicalPath())
    except api.exc.CannotGetPortalError:
        site_path = '(zoperoot)'
    transaction.get().addAfterCommitHook(record, args=(recorder, site_path))



import json
import logging
import time
import unittest

from plone import api
from plone.app.testing import (TEST_USER_ID, TEST_USER_NAME, login, setRoles)
from Products.CMFCore.utils import getToolByName
import transaction
from zope.annotation.interfaces import IAnnotations

from castle.cms.browser.search import SearchAjax
from castle.cms.indexing import hps
from castle.cms.social import COUNT_ANNOTATION_KEY
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from castle.cms.tests.utils import get_tile


logger = logging.getLogger(__name__)


class TestHPS(unittest.TestCase):
    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))

        if not self._is_enabled():
            return

        transaction.begin()
        folder = api.content.create(
            type='Folder',
            id='folder1',
            container=self.portal,
            title='Foobar folder')
        folder2 = api.content.create(
            type='Folder',
            id='folder2',
            container=folder,
            title='Foobar subfolder')
        doc1 = api.content.create(
            type='Document',
            id='doc1',
            container=folder,
            title='Foobar one')
        doc2 = api.content.create(
            type='Document',
            id='doc2',
            container=folder,
            subject=('foobar',),
            title='Foobar two')
        doc3 = api.content.create(
            type='Document',
            id='doc3',
            container=folder,
            title='Foobar three')
        doc4 = api.content.create(
            type='Document',
            id='doc4',
            container=folder,
            title='Foobar four')
        doc5 = api.content.create(
            type='Document',
            id='doc5',
            container=folder2,
            title='Foobar five')
        doc6 = api.content.create(
            type='Document',
            id='doc6',
            container=folder,
            title='Foobar six',
            exclude_from_search=True)
        ann = IAnnotations(doc2)
        ann[COUNT_ANNOTATION_KEY] = {
            'twitter_matomo': 5,
            'facebook': 5,
        }
        for item in [folder, doc1, doc2, doc3, doc5]:
            api.content.transition(obj=item, to_state='published')
            item.reindexObject()

        self._hps_update()
        transaction.commit()

        hpsconn = hps.get_connection()
        hpsconn.indices.flush(index=hps.get_index_name())

    def _is_enabled(self):
        if not hps.is_enabled() or not hps.health_is_good():
            logger.warn("disabling hps testing because it's not enabled or there is a connection issue")
            return False
        return True

    def _hps_update(self):
        catalog = getToolByName(self.portal, 'portal_catalog')
        hpscat = hps.get_catalog()
        hpscat.recreateCatalog()
        catalog.manage_catalogRebuild()

    def tearDown(self):
        if not self._is_enabled():
            return

        transaction.begin()
        api.content.delete(self.portal.folder1)
        transaction.commit()

    def _test_ajax_search_rank_social(self):
        if not self._is_enabled():
            return

        self.request.form.update({
            'SearchableText': 'Foobar',
            'portal_type': 'Document'
        })
        view = SearchAjax(self.portal, self.request)
        result = json.loads(view())
        self.assertEquals(result['count'], 3)
        self.assertEquals(result['results'][0]['path'], '/folder1/doc2')

    def test_ajax_search_pt(self):
        if not self._is_enabled():
            return

        time.sleep(1)
        self.request.form.update({
            'SearchableText': 'Foobar',
            'portal_type': 'Folder'
        })
        view = SearchAjax(self.portal, self.request)
        result = json.loads(view())
        self.assertEquals(result['count'], 1)
        self.assertEquals(result['results'][0]['path'], '/folder1')

    def test_ajax_search_subject(self):
        if not self._is_enabled():
            return

        time.sleep(1)
        self.request.form.update({
            'SearchableText': 'Foobar',
            'Subject': 'foobar'
        })
        view = SearchAjax(self.portal, self.request)
        result = json.loads(view())
        self.assertEquals(result['count'], 1)
        self.assertEquals(result['results'][0]['path'], '/folder1/doc2')

    def test_es_querylisting_unicode_issue(self):
        if not self._is_enabled():
            return

        tile = get_tile(self.request, self.portal, 'castle.cms.querylisting', {})
        # should not cause errors...
        self.request.form.update({
            'Title': 'ma\xf1on'
        })
        self.assertTrue(tile.filter_pattern_config != '{}')
        tile()

    def test_ajax_search_with_private_parents(self):
        if not self._is_enabled():
            return

        self.request.form.update({
            'SearchableText': 'Foobar',
            # 'Subject': 'foobar'
        })
        time.sleep(3)  # wait a little to let the search catch up?
        view_1 = SearchAjax(self.portal, self.request)
        result_1 = json.loads(view_1())
        self.assertEqual(result_1['count'], 4)
        api.portal.set_registry_record('plone.allow_public_in_private_container', True)
        view_2 = SearchAjax(self.portal, self.request)
        result_2 = json.loads(view_2())
        self.assertEqual(result_2['count'], 5)



import logging

# from opensearch import helpers
from opensearchpy import TransportError
from plone import api
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import _getAuthenticatedUser
from urlparse import urljoin
from urlparse import urlparse
from wildcard.hps.interfaces import IQueryAssembler
from wildcard.hps.opensearch import WildcardHPSCatalog
from zope.component import getMultiAdapter
from zope.component import getUtility


logger = logging.getLogger("Plone")


def get_catalog():
    catalog = api.portal.get_tool('portal_catalog')
    hps = WildcardHPSCatalog(catalog)
    return hps


def get_connection():
    hpscat = get_catalog()
    return hpscat.connection


def get_index_name():
    hpscat = get_catalog()
    return hpscat.index_name


def get_bulk_size():
    hpscat = get_catalog()
    return hpscat.get_setting("bulk_size", 50)


def is_enabled():
    hpscat = get_catalog()
    return hpscat.enabled


def health_is_good():
    conn = get_connection()
    try:
        return conn.cluster.health()['status'] in ('green', 'yellow')
    except Exception:
        return False


def hps_get_number_of_matches(index_name, query):
    conn = get_connection()
    results = conn.count(index=index_name, body=query)
    return results.get('count', -1)


def hps_get_data(index_name, query, **kwargs):
    conn = get_connection()
    results = conn.search(index=index_name, body=query, **kwargs)
    try:
        data = results['hits']['hits']
        total = results['hits']['total']['value']
        scroll_id = results.get('_scroll_id', None)
        return data, total, scroll_id
    except KeyError:
        return [], -1, None


def hps_get_scroll(scroll_id, **kwargs):
    conn = get_connection()
    results = conn.scroll(scroll_id=scroll_id, **kwargs)
    try:
        data = results['hits']['hits']
        total = results['hits']['total']['value']
        scroll_id = results.get('_scroll_id', None)
        return data, total, scroll_id
    except KeyError:
        return [], -1, None


def gen_audit_query(
        field_map_prefix=None, sort=None, sortdir=None,
        schema_version="1", schema_type="castle.cms.audit", instance=None, site=None,
        typeval=None, user=None, content=None, after=None, before=None):

    sortfield = sort
    schemaversionfield = "schema_version"
    schematypefield = "schema_type"
    instancefield = "instance"
    sitefield = "site"
    typefield = "type"
    userfield = "user"
    objectfield = "object"
    datefield = "date"
    if field_map_prefix is not None:
        sortfield = "{}.{}".format(field_map_prefix, sortfield)
        schemaversionfield = "{}.{}".format(field_map_prefix, schemaversionfield)
        schematypefield = "{}.{}".format(field_map_prefix, schematypefield)
        instancefield = "{}.{}".format(field_map_prefix, instancefield)
        sitefield = "{}.{}".format(field_map_prefix, sitefield)
        typefield = "{}.{}".format(field_map_prefix, typefield)
        userfield = "{}.{}".format(field_map_prefix, userfield)
        objectfield = "{}.{}".format(field_map_prefix, objectfield)
        datefield = "{}.{}".format(field_map_prefix, datefield)

    filters = [
        {'match_phrase': {schemaversionfield: schema_version}},
        {'match_phrase': {schematypefield: schema_type}},
        {'match_phrase': {instancefield: instance}},
        {'match_phrase': {sitefield: site}},
    ]

    if typeval is not None and len(typeval) > 0:
        filters.append({'match_phrase': {typefield: typeval}})

    if user is not None and len(user) > 0:
        filters.append({'match_phrase': {userfield: user}})

    if content is not None and len(content) > 0:
        items = content.split(';')
        if len(items) > 0:
            filters.append({'terms': {objectfield: items}})

    if (after is not None and len(after) > 0) or (before is not None and len(before) > 0):
        rangefilter = {'range': { datefield: {}}}
        if after is not None:
            rangefilter['range'][datefield]['gte'] = after
        if before is not None:
            rangefilter['range'][datefield]['lte'] = before
        rangefilter['range'][datefield]['format'] = "strict_date_optional_time"
        filters.append(rangefilter)

    if len(filters) == 0:
        query = {"query": {'match_all': {}}}
    else:
        filters.append({"match_all": {}})
        query = {
            "query": {
                'bool': {
                    'filter': filters
                }
            }
        }

    if sort is not None:
        sortdir = sortdir or "desc"
        sortquery = [{sortfield: {"order": sortdir, "unmapped_type": "boolean"}}]
        query["sort"] = sortquery

    return query


def index_in_es(obj):
    catalog = api.portal.get_tool('portal_catalog')
    es = ElasticSearchCatalog(catalog)
    if es.enabled:
        index_batch([], {IUUID(obj): obj}, [], es)


def get_index_summary(index_name, terms, agg_agg_terms=None, filter_terms=None):
    query = {
        "size": 0,
        "aggregations": {
            "totals": {
                "terms": terms
            }
        }
    }
    if agg_agg_terms is not None:
        query["aggregations"]["totals"]["aggregations"] = {
            "types": {
                "terms": agg_agg_terms
            }
        }
    if filter_terms is not None:
        query["bool"] = {
            "filter": [{"term": filter_terms}]
        }
    try:
        conn = get_connection()
        result = conn.search(index=index_name, body=query)
    except TransportError:
        logger.error("problem getting index summary for {}".format(index_name), exc_info=True)
        return []

    data = result['aggregations']['totals']['buckets']
    return data


def get_search_results(context, request, catalog, search_attributes, page, page_size, query):
    """
    page: int
    page_size: int, max 10000 (opensearch limit)
    query: dict containing key/value pairs for filters/search fields
    """

    # start is used to skip a number of records returned for the query from
    # opensearch. We don't keep a scroll here, or some other mechanism, to make
    # sure the results are 100% the same between results. This is intentional,
    # as we don't mind if the results and pages change a little between searches
    start = (page - 1) * page_size

    site_path = '/'.join(context.getPhysicalPath())
    base_site_url = context.absolute_url()

    return_data = {
        'count': 0,
        'results': [],
        'page': page,
        'suggestions': [],
    }

    # no query parameters means no results
    if len(query) <= 0:
        return return_data

    user = _getAuthenticatedUser(catalog)
    query['allowedRolesAndUsers'] = catalog._listAllowedRolesAndUsers(user)

    # convert the query parameter into a usable opensearch query
    catalog = api.portal.get_tool('portal_catalog')
    hps = WildcardHPSCatalog(catalog)
    index_name = hps.index_name
    query_assembler = getMultiAdapter((request, hps), IQueryAssembler)
    normalized_query, sort_list = query_assembler.normalize(query)
    os_query = query_assembler(normalized_query)

    # if a specific site's crawl data is being queried, exclude non-applicable
    # results from the generated opensearch query
    if 'searchSite' in request.form:
        index_name = '{}_crawler'.format(index_name=hps.index_name)
        os_query = os_query['script_score']['query']
        os_query['bool']['filter'] = [
            {
                'term': {
                    'domain': request.form['searchSite'],
                },
            },
        ]

    # final query includes suggestions on SearchableText and specifying a sort
    final_query = {
        'query': os_query,
        'suggest': {
            'SearchableText': {
                'text': query.get('SearchableText', ''),
                'term': {
                    'field': 'SearchableText',
                },
            },
        },
        'sort': sort_list
    }

    # these are parameters to the search endpoint, but not part of the query
    # itself
    os_query_params = {
        'stored_fields': ','.join(search_attributes),
        'from_': start,
        'size': page_size,
    }

    os_results = hps.connection.search(
        index=index_name,
        body=final_query,
        **os_query_params)

    return_data['count'] = os_results.get('hits', {}).get('total', {}).get('value', 0)

    try:
        return_data['suggestions'] = os_results['suggest']['SearchableText'][0]['options']
    except Exception:
        # likely an index or key error, which means no
        # suggestions, which means we already have an empty list
        pass

    # view_types are just types that should have '/view' appended to the path
    # of the url associated with the object
    registry = getUtility(IRegistry)
    view_types = registry.get('plone.types_use_view_action_in_listings', [])

    data = os_results.get('hits', {}).get('hits', {})
    for result in data:
        fields = result['fields']

        # build an item from the current result datum
        item = {}

        # we only care about the given attributes
        for key in search_attributes:
            val = fields.get(key, None)
            # if a value is a list, but it only has one element, reduce it
            # to just the element value
            if val is not None and type(val) == list and len(val) == 1:
                val = val[0]
            item[key] = val

        # either build or parse out the url, base_url, and path for the object
        if 'url' in fields:
            url = fields['url']
            if url is not None and type(url) == list and len(url) == 1:
                url = url[0]

            base_url = url

            parsed_url = urlparse(url)
            path = parsed_url.path
        else:
            path = fields.get('path.path', '')
            if path is not None and type(path) == list and len(path) == 1:
                path = path[0]
            path = path[len(site_path):]
            url = urljoin('{}/'.format(base_site_url), path.lstrip('/'))

            base_url = url

            portal_type = item.get('portal_type', None)
            if portal_type is not None and portal_type in view_types:
                url = '{}/view'.format(url)

        item['review_state'] = 'published'
        item['score'] = result['_score']
        item['url'] = url
        item['base_url'] = base_url
        item['path'] = path

        return_data['results'].append(item)

    return return_data



from plone.app.registry.browser import controlpanel
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from castle.cms import cache
from castle.cms.constants import CRAWLED_DATA_KEY
from castle.cms.indexing import hps
from castle.cms.indexing import crawler
from castle.cms.interfaces import ICrawlerConfiguration


class CrawlerControlPanelForm(controlpanel.RegistryEditForm):
    schema_prefix = 'castle'
    schema = ICrawlerConfiguration
    id = "CrawlerControlPanel"
    label = u"Site Crawler Configuration"
    description = "Configure CastleCMS to crawl other sites and include " \
                  "those results in your site search. WildcardHPS must " \
                  "be enabled."


class CrawlerControlPanel(controlpanel.ControlPanelFormWrapper):
    form = CrawlerControlPanelForm
    index = ViewPageTemplateFile('templates/crawler.pt')

    def get_crawl_data(self):
        try:
            data = cache.get_client(CRAWLED_DATA_KEY)
        except Exception:
            return {}

        return data

    def get_index_summary(self):
        idx = crawler.index_name(hps.get_index_name())
        terms = dict(field="domain")
        result = hps.get_index_summary(idx, terms)
        return result



import argparse
import gzip
import logging
import sys
import time

from DateTime import DateTime
from lxml import etree
from lxml import html
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
import requests
from StringIO import StringIO
from tendo import singleton
from urlparse import urlparse
from zope.component import getUtility

from castle.cms import archival
from castle.cms import cache
from castle.cms.constants import CRAWLED_DATA_KEY
from castle.cms.cron.utils import login_as_admin
from castle.cms.cron.utils import setup_site
from castle.cms.cron.utils import spoof_request
from castle.cms.files import aws
from castle.cms.indexing import hps
from castle.cms.indexing import crawler as hpscrawl
from castle.cms.interfaces import ICrawlerConfiguration
from castle.cms.utils import clear_object_cache


# override normal plone logging and use a configured root logger here
# to capture output to stdout since this is a script that will need to render
# output of logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


class Crawler(object):
    _meta_properties = {
        'Title': [
            'meta[property="og:title"]',
            'title',
            '.documentFirstHeading',
            'h1'
        ],
        'Description': [
            'meta[property="og:description"]',
        ],
        'created': [
            'meta[name="DC.date.created"]',
            'meta[name="creationDate"]',
            'meta[name="publicationDate"]',
        ],
        'modified': [
            'meta[name="DC.date.modified"]',
            'meta[name="modificationDate"]',
            'meta[name="publicationDate"]',
        ],
        'effective': [
            'meta[name="publicationDate"]',
        ],
        'portal_type': [
            'meta[name="DC.type"]',
            'meta[name="portalType"]',
            'meta[property="og:type"]',
        ],
        'image_url': [
            'meta[property="og:image"]',
        ]
    }
    date_fields = ['created', 'modified', 'effective']
    searchable_text_selector = ','.join([
        '#content p,#content h2,#content h3,#content ul li'
    ])

    def __init__(self, site, settings, hpscatalog):
        self.site = site
        self.settings = settings
        self.hpscatalog = hpscatalog
        self.index_name = hpscrawl.index_name(hps.get_index_name())

        self.site._p_jar.sync()  # sync transactions

        self.data = cache.get_client(CRAWLED_DATA_KEY)

    def crawl_page(self, url):
        logger.info('Indexing ' + url)
        try:
            resp = requests.get(url, headers={
                'User-Agent': self.settings.crawler_user_agent
            })
        except Exception:
            # unable to access the page, remove for now
            return False
        if resp.status_code == 404 or 'html' not in resp.headers.get('content-type', ''):
            # remove from index
            return False
        try:
            dom = html.fromstring(resp.content)
        except etree.XMLSyntaxError:
            # unable to parse html, remove for now
            return False  # lxml has been known to throw this as a bug, maybe use BeautifulSoup
        parsed = urlparse(url)
        data = {
            'url': url,
            'domain': parsed.netloc
        }

        for name, selectors in self._meta_properties.items():
            for selector in selectors:
                result = dom.cssselect(selector)
                if len(result) > 0:
                    result = result[0]
                    if result.attrib.get('content'):
                        data[name] = result.attrib['content']
                        break
                    elif result.text_content().strip():
                        data[name] = result.text.strip()
                        break

        for date_field in self.date_fields:
            val = data.get(date_field)
            if val:
                try:
                    data[date_field] = DateTime(val).ISO8601()
                except Exception:
                    pass

        searchable_text = [
            data.get('Title', ''),
            data.get('Description', '')
        ]
        for el in dom.cssselect(self.searchable_text_selector):
            searchable_text.append(el.text_content())
        data['SearchableText'] = ' '.join(searchable_text)

        return data

    def crawl_archive_url(self, url):
        # archive urls don't need to be reindexed ever, they
        # are in a RO archive
        if hpscrawl.url_is_indexed(self.hpscatalog, self.index_name, url):
            return

        data = self.crawl_page(url)
        if not data:
            return

        data['sitemap'] = 'archives'
        hpscrawl.index_doc(self.hpscatalog, self.index_name, url, data)

    def crawl_archives(self):
        registry = getUtility(IRegistry)
        base_url = registry.get('castle.aws_s3_base_url', None)

        storage = archival.Storage(self.site)
        urls = []
        for _, archive_data in storage.archives.items():
            # archives never need to be re-indexed
            url = archive_data.get('view_url', None) or archive_data['url']
            urls.append(aws.swap_url(url, base_url=base_url))

        existing_urls = hpscrawl.get_all_ids(self.hpscatalog, self.index_name, "archives")
        for _id in set(urls) - set(existing_urls):
            # pages that have not yet been crawled
            try:
                self.crawl_archive_url(_id)
            except Exception:
                logger.error('Error indexing archive url: ' + _id, exc_info=True)

        for _id in set(existing_urls) - set(urls):
            # pages that have been removed from the archive
            hpscrawl.delete_from_index(self.hpscatalog, self.index_name, _id)

    def clean_removed_pages(self, sitemap, crawled_urls):
        parsed = urlparse(sitemap)
        domain = parsed.netloc
        ids = hpscrawl.get_all_ids(self.hpscatalog, self.index_name, domain)
        for _id in set(ids) - set(crawled_urls):
            hpscrawl.delete_from_index(self.hpscatalog, self.index_name, _id)

    def crawl_site_map(self, sitemap, full=False):
        resp = requests.get(sitemap, headers={
            'User-Agent': self.settings.crawler_user_agent
        })
        if resp.status_code != 200:
            logger.error('Not a valid sitemap response for %s' % sitemap)
            return

        self.site._p_jar.sync()  # sync transactions

        try:
            last_crawled = DateTime(self.data[sitemap])
        except Exception:
            # KeyError or date parsing issue just revert to old time
            last_crawled = DateTime('1999/01/01')

        try:
            self.data[sitemap] = DateTime().ISO8601().decode('utf8')
        except Exception:
            # maybe a storage error or something -- we'll just not set a new time
            # if that happens, and the crawler should pick up crawling the object
            # again
            pass

        clear_object_cache(self.site)

        if sitemap.lower().endswith('.gz'):
            sitemap_content = gzip.GzipFile(fileobj=StringIO(resp.content)).read()
        else:
            sitemap_content = resp.content

        dom = etree.fromstring(sitemap_content)
        crawled_urls = []
        for url_node in dom.xpath("//*[local-name() = 'url']"):
            loc = url_node.xpath("*[local-name() = 'loc']")
            if loc:
                loc = loc[0].text.strip()
            else:
                loc = None
            url = loc
            crawled_urls.append(url)

            lastmod = url_node.xpath("*[local-name() = 'lastmod']")
            if lastmod:
                lastmod = lastmod[0].text.strip()
            else:
                lastmod = None
            if lastmod:
                lastmod = DateTime(lastmod)
                if not full and lastmod < last_crawled:
                    continue

            if not url:
                continue
            try:
                interval = self.settings.crawler_interval
            except Exception:
                interval = 0
            time.sleep(interval)
            data = self.crawl_page(url)
            if data is False:
                crawled_urls.remove(url)
                hpscrawl.remove_doc_from_index(self.hpscatalog, self.index_name, url)
            else:
                data['sitemap'] = sitemap
                hpscrawl.index_doc(self.hpscatalog, self.index_name, url, data)
                crawled_urls.append(url)

        self.clean_removed_pages(sitemap, crawled_urls)


def crawl_site(site, full=False):
    registry = getUtility(IRegistry)
    settings = registry.forInterface(ICrawlerConfiguration, prefix='castle')
    if not settings.crawler_active or not settings.crawler_site_maps:
        logger.info("Crawler must first be enabled in Site Setup")
        return False

    if not hps.is_enabled():
        logger.info("WildcardHPS must be enabled in Site Setup to use Crawler")
        return False

    hpscatalog = hps.get_catalog()
    hpscrawl.ensure_index_exists(hpscatalog, hpscrawl.index_name(hps.get_index_name()))
    crawler = Crawler(site, settings, hpscatalog)

    if settings.crawler_index_archive:
        crawler.crawl_archives()

    for sitemap in settings.crawler_site_maps:
        try:
            crawler.crawl_site_map(sitemap, full)
        except Exception:
            logger.error('Error crawling site map: %s' % sitemap, exc_info=True)

    return True


def run(app):
    singleton.SingleInstance('crawler')

    app = spoof_request(app)  # noqa
    login_as_admin(app)  # noqa

    parser = argparse.ArgumentParser(
        description="Index configured sites, archives, and sitemaps for high performance search")
    parser.add_argument('--site-id', dest='site_id', default=None)
    parser.add_argument('--partial', dest='partial', action='store_false')

    args, _ = parser.parse_known_args()

    try:
        if args.site_id is not None:
            setup_site(app[args.site_id])
            crawl_site(app[args.site_id])
        else:
            for oid in app.objectIds():
                obj = app[oid]
                if IPloneSiteRoot.providedBy(obj):
                    try:
                        setup_site(obj)
                        obj._p_jar.sync()  # sync transactions
                        crawl_site(obj, not args.partial)
                    except Exception:
                        logger.error('Error crawling site %s' % oid, exc_info=True)
    except KeyError:
        pass
    except Exception:
        logger.error('Error setting up crawling', exc_info=True)


if __name__ == '__main__':
    run(app)  # noqa



from pkg_resources import get_distribution


def get_castle_version(strip_dev=False):
    version_number = get_distribution('castle.cms').version
    if strip_dev:
        version_parts = version_number.split('.')
        is_dev_version = version_parts[-1].find('dev') == 0
        return_version_parts = version_parts[:-1 if is_dev_version else None]
        return '.'.join(return_version_parts)
    return version_number


CASTLE_VERSION = get_castle_version()
CASTLE_VERSION_STRING = 'CastleCMS {}'.format(CASTLE_VERSION)



from Products.CMFPlone.resources import add_resource_on_request
from castle.cms.browser.audit import AuditView as BaseAuditView


class AuditView(BaseAuditView):
    def __call__(self):
        # utility function to add resource to rendered page
        add_resource_on_request(self.request, 'castle-components-audit')
        return super(AuditView, self).__call__()

    label = 'Audit Log'
    user = False



from castle.cms import utils
from castle.cms.interfaces import ITemplate
from plone import api
from zope.component.hooks import getSite
from zope.interface import alsoProvides


def save_as_template(obj, action):
    site = getSite()
    folder = utils.recursive_create_path(site, '/template-repository')
    folder.exclude_from_nav = True
    if action == 'convert':
        template_obj = api.content.move(source=obj, target=folder)
    elif action == 'copy':
        template_obj = api.content.copy(source=obj, target=folder)

    alsoProvides(template_obj, ITemplate)

    try:
        site.template_list.append(template_obj)
    except AttributeError:
        site.template_list = [template_obj]

    return template_obj



"""
The data submitted to a GELF endpoint logger for this is using a schema that
looks something like this:

class UserAccess(BaseModel):
    schema_version: str
    schema_type: str
    app_name: str
    site: str
    report_id: str
    username: str
    group: str
    permissions: list[str]

In this case, Plone (by way of Zope and AccessControl) assigns *permissions*
to *roles*, and *roles* to *users* in a 3-layer approach.

Therefore, in the above model, the following are mapped to an individual record:

    - username == user.getUserName()
    - group == a single role
    - permissions == all permissions associated with the Role that are active for the user

    - schema_version == SCHEMA_VERSION variable
    - schema_type == SCHEMA_TYPE variable
    - app_name == APP_NAME variable
    - site == app[site].id value
    - report_id == UUID4 unique per-site and per-execution of this script
"""
from argparse import ArgumentParser
import csv
from datetime import datetime
import logging
import os
import sys
import uuid

from AccessControl.SecurityManagement import newSecurityManager
from plone import api
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from tendo import singleton
from zope.component.hooks import setSite

from castle.cms.gelf import GELFHandler



logger = logging.getLogger("Plone")


SCHEMA_VERSION = "1"
SCHEMA_TYPE = "UserGroupMatrix"
APP_NAME = "castle.cms"
gelflogger = logging.getLogger(SCHEMA_TYPE)
gelfhandler = GELFHandler()
gelfformatter = logging.Formatter("%(asctime)s %(name)s [%(levelname)s] %(message)s")
gelfhandler.setFormatter(gelfformatter)
gelflogger.addHandler(gelfhandler)


def get_args():
    parser = ArgumentParser(
        description='Get a report of permissions by role and user')
    parser.add_argument('--site-id', dest='site_id', default='Castle')
    args, _ = parser.parse_known_args()
    return args


def report_on_users(site):
    extras = {
        "schema_version": SCHEMA_VERSION,
        "schema_type": SCHEMA_TYPE,
        "app_name": APP_NAME,
        "site": site.id,
        "report_id": uuid.uuid4(),
    }

    users = api.user.get_users()
    for user in users:
        extras["username"] = user.getUserName()

        perms = api.user.get_permissions(user=user)
        roles = api.user.get_roles(user=user)
        roleperms = {}

        # roles is a list of names of roles the user has
        # perms is a dict of permname:bool of what the user has
        # go through each role the user has, then collect the permissions associated with that
        #   specific role by iterating through roles related to each permission the user has.
        #
        #   I know what you're thinking, but yes, this is how it needs to be done (for now).
        #   the upstream packages don't have a convenient way to just get a list of
        #   permissions associated with a given role. They either give all permissions or roles
        #   associated with a user, or get all roles associated with a permission.
        #   Useful in most contexts outside of generating a report like this.
        for role in roles:
            roleperms[role] = []
            for perm in perms.keys():
                if perms[perm]:
                    roleperms[role].append(perm)

        for rp in roleperms.keys():
            msg = "{schemaversion} {schematype} {appname} {site} {reportid} {username} {rolename} {permissions}".format(
                schemaversion=extras["schema_version"],
                schematype=extras["schema_type"],
                appname=extras["app_name"],
                site=extras["site"],
                reportid=extras["report_id"],
                username=extras["username"],
                rolename=rp,
                permissions=",".join(roleperms[rp]))
            extras["group"] = rp
            extras["permissions"] = roleperms[rp]
            gelflogger.info(msg, extras=extras)


def run(app):
    singleton.SingleInstance('reportusers')

    args = get_args()

    user = app.acl_users.getUser('admin')
    newSecurityManager(None, user.__of__(app.acl_users))

    if args.site_id.strip().lower() == "_all_":
        for oid in app.objectIds():
            obj = app[oid]
            if IPloneSiteRoot.providedBy(obj):
                setSite(obj)
                report_on_users(obj)
    else:
        site = app[args.site_id]
        setSite(site)
        report_on_users(site)


def setup_and_run():
    conf_path = os.getenv("ZOPE_CONF_PATH", "parts/instance/zope.conf")
    if conf_path is None or not os.path.exists(conf_path):
        raise Exception('Could not find zope.conf at {}'.format(conf_path))

    from Zope2 import configure
    configure(conf_path)
    import Zope2
    app = Zope2.app()
    from Testing.ZopeTestCase.utils import makerequest
    app = makerequest(app)
    app.REQUEST['PARENTS'] = [app]
    from zope.globalrequest import setRequest
    setRequest(app.REQUEST)
    from AccessControl.SpecialUsers import system as user
    from AccessControl.SecurityManagement import newSecurityManager
    newSecurityManager(None, user)

    run(app)


if __name__ == '__main__':
    run(app)  # noqa: F821



import argparse
import csv
import datetime
import logging
import sys

from AccessControl.SecurityManagement import newSecurityManager
from tendo import singleton
from zope.component.hooks import setSite

from castle.cms import audit
from castle.cms.indexing import hps


logger = logging.getLogger("castle.cms")
logger.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


def get_args():
    parser = argparse.ArgumentParser(
        description='Export Audit Logs to CSV for a specific site')
    parser.add_argument(
        '--site-id', dest='site_id', default=None,
        help='site id to fetch audit log for')
    parser.add_argument(
        '--filepath', dest='filepath', default='./exported-audit-data.csv',
        help='where to export data in csv format')
    parser.add_argument(
        '--scrolltime', dest='scrolltime', default='2s',
        help='ES scroll time, scroll api is used for ES2.3 compat')
    parser.add_argument('--searchtimeout', dest='searchtimeout', default='10s')
    parser.add_argument('--timeout', dest='timeout', type=float, default=0.5)
    args, _ = parser.parse_known_args()
    return args


def convertunicode(s):
    if isinstance(s, unicode):  # noqa: F821
        return s.encode('utf-8')
    return s


def export(args):
    logger.info("exporting from index `{}`".format(audit.get_index_name()))
    starttime = datetime.datetime.now()

    logger.info("fetching resultset with scroll time of `{}`".format(args.scrolltime))

    query = {"query": {'match_all': {}}}
    size = hps.hps_get_number_of_matches(audit.get_index_name(), query)
    if size < 0:
        logger.info('unknown number of items to export')
    else:
        logger.info("{} items need to be exported".format(size))

    results, _, scroll_id = hps.hps_get_data(
        audit.get_index_name(),
        query,
        sort='date:desc',
        scroll=args.scrolltime,
        size=10000, # max per search result
        timeout=args.searchtimeout)

    logger.info("writing to `{}` (truncated)".format(args.filepath))
    with open(args.filepath, 'w') as output:
        writer = csv.writer(output, quoting=csv.QUOTE_ALL)
        logger.info("writing header row...")
        writer.writerow(['date', 'name', 'object', 'path', 'request_uri', 'summary', 'type', 'user'])
        num = 0
        while len(results) > 0:
            old_scroll_id = scroll_id
            logger.info("writing {} hits for scroll {}".format(len(results), old_scroll_id))
            for result in results:
                data = result['_source']
                rowdata = [
                    data.get("date", ""),
                    data.get("name", ""),
                    data.get("object", ""),
                    data.get("path", ""),
                    data.get("request_uri", ""),
                    data.get("summary", ""),
                    data.get("type", ""),
                    data.get("user", ""),
                ]
                rowdata = [convertunicode(a) for a in rowdata]
                writer.writerow(rowdata)
            num += len(results)
            logger.info("{} of {} written".format(num, size))
            logger.info("fetching next scroll...")
            results, _, old_scroll_id = hps.hps_get_scroll(scroll_id=old_scroll_id, scroll=args.scrolltime)

    endtime = datetime.datetime.now()
    deltatime = endtime - starttime
    logger.info("export complete -- took {}s, exported {} records".format(deltatime.total_seconds(), num))


def run(app):
    singleton.SingleInstance('exportauditlog')

    args = get_args()

    user = app.acl_users.getUser('admin')  # noqa: F821
    newSecurityManager(None, user.__of__(app.acl_users))  # noqa: F821
    if args.site_id is not None:
        site = app[args.site_id]
        setSite(site)

    export(args)


if __name__ == '__main__':
    run(app)  # noqa: F821



from castle.cms.widgets import SelectFieldWidget
from plone.autoform import directives
from plone.autoform.interfaces import IFormFieldProvider
from plone.supermodel import model
from zope import schema
from zope.interface import alsoProvides


class ISearch(model.Schema):

    model.fieldset(
        'settings',
        fields=[
            'searchterm_pins',
            'robot_configuration',
            'backend_robot_configuration',
            'exclude_from_search',
        ],
    )

    searchterm_pins = schema.List(
        title=u'Search Term Pins',
        description=u'Searches this content should show top in results for.',
        required=False,
        value_type=schema.TextLine(),
        default=[]
    )

    directives.widget('robot_configuration', SelectFieldWidget)
    robot_configuration = schema.List(
        title=u'How robots should behave with this content on frontend urls',
        description=u'Robots index and consume content on the web. '
                    u'This allows you to remove content from search indexes.',
        required=False,
        defaultFactory=lambda: ['index', 'follow'],
        missing_value=[],
        value_type=schema.Choice(
            vocabulary='castle.cms.vocabularies.RobotBehaviors',
        ),
    )

    directives.widget('backend_robot_configuration', SelectFieldWidget)
    backend_robot_configuration = schema.List(
        title=u'How robots should behave with this content on backend urls',
        description=u'Robots index and consume content on the web. '
                    u'This allows you to remove content from search indexes.',
        required=False,
        defaultFactory=lambda: [],
        missing_value=[],
        value_type=schema.Choice(
            vocabulary='castle.cms.vocabularies.RobotBehaviors',
        )
    )

    exclude_from_search = schema.Bool(
        title=u'Exclude from searches',
        description=u'If selected, this item will not appear in searches',
        required=True,
        default=False
    )


alsoProvides(ISearch, IFormFieldProvider)



from plone.app.versioningbehavior import _
from plone.app.versioningbehavior.behaviors import IVersionable as OldIVersionable
from plone.autoform.interfaces import IFormFieldProvider
from zope import schema
from zope.interface import alsoProvides
from zope.interface import Interface


class IVersionable(OldIVersionable):

    changeNote = schema.TextLine(
        title=_(u'label_change_note', default=u'Change Note'),
        description=_(u'help_change_note',
                      default=u'Enter a comment that describes the changes you made.'),
        required=True)


alsoProvides(IVersionable, IFormFieldProvider)


class IVersioningSupport(Interface):
    """
    Marker Interface for the IVersionable behavior.
    """



from plone.autoform.interfaces import IFormFieldProvider
from plone.dexterity.interfaces import IDexterityContent
from plone.supermodel import model
from zope.component import adapter
from zope.interface import implementer
from zope.interface import provider
from castle.cms.constants import (
    DEFAULT_FONT_SIZE_SMALL,
    DEFAULT_FONT_SIZE_MEDIUM,
    DEFAULT_FONT_SIZE_LARGE,
    VALID_CSS_FONT_SIZE_PATTERN,
)
from plone.app.contenttypes import _
from plone.supermodel import directives
from zope.schema import Choice
from zope.schema import TextLine
from zope.interface import invariant
from zope.interface import Invalid
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary
from plone.api.portal import get_registry_record


adjustable_font_size_fieldset_description = _(
    u'''Use these settings to adjust the font size for the specified tiles on this content item.
Default settings are small: {}, medium: {}, large: {}, but those can be changed in Castle Settings.
Absolute sizes such as 'px' and 'pt' can be used.  Ex: 15px'''.format(
        DEFAULT_FONT_SIZE_SMALL,
        DEFAULT_FONT_SIZE_MEDIUM,
        DEFAULT_FONT_SIZE_LARGE,
    )
)


def get_inline_style(context, tile_type):
    sizes = {
        'default': None,
        'small': get_registry_record(
            'castle.font_size_small',
            default=DEFAULT_FONT_SIZE_SMALL,
        ),
        'medium': get_registry_record(
            'castle.font_size_medium',
            default=DEFAULT_FONT_SIZE_MEDIUM,
        ),
        'large': get_registry_record(
            'castle.font_size_large',
            default=DEFAULT_FONT_SIZE_LARGE,
        ),
        'custom': getattr(
            context,
            'font_size_custom_{}'.format(tile_type),
            None,
        )
    }
    font_size_choice_value = getattr(
        context,
        'font_size_choice_{}'.format(tile_type),
        'default',
    )
    custom_font_size = sizes[font_size_choice_value]
    return (
        '' if not custom_font_size
        else 'font-size: {};'.format(custom_font_size)
    )


def font_size_constraint(provided_font_size):
    return bool(VALID_CSS_FONT_SIZE_PATTERN.match(provided_font_size))


@provider(IFormFieldProvider)
class IAdjustableFontSizeQueryListing(model.Schema):
    directives.fieldset(
        'font_size_adjustments',
        description=adjustable_font_size_fieldset_description,
        label=_(u'Font Size Adjustments'),
        fields=(
            'font_size_choice_query_listing',
            'font_size_custom_query_listing',
        ),
    )
    font_size_choice_query_listing = Choice(
        title=_(u'Font size for Query Listing tile'),
        vocabulary=SimpleVocabulary([
            SimpleTerm('default', 'default', u'Default'),
            SimpleTerm('small', 'small', u'Small'),
            SimpleTerm('medium', 'medium', u'Medium'),
            SimpleTerm('large', 'large', u'Large'),
            SimpleTerm('custom', 'custom', u'Custom'),
        ]),
        default='default',
        required=True,
    )
    font_size_custom_query_listing = TextLine(
        title=u'Custom font size for Query Listing tile',
        description=_(u'Only used if "Custom" is selected immediately above'),
        required=False,
        constraint=font_size_constraint,
    )

    @invariant
    def verify_custom_font_size_specified_if_custom(data):
        font_size_choice_value = getattr(data, 'font_size_choice_query_listing')
        font_size_custom_value = getattr(data, 'font_size_custom_query_listing')
        if font_size_choice_value == 'custom' and not font_size_custom_value:
            raise Invalid(_(
                u'Custom font size must be provided for Query Listing tile if "Custom" is selected choice'
            ))


@implementer(IAdjustableFontSizeQueryListing)
@adapter(IDexterityContent)
class AdjustableFontSizeQueryListing(object):

    def __init__(self, context):
        self.context = context



import json
import logging
import os
import socket
import time

import requests


logger = logging.getLogger("Plone")

_dummy_log_record = logging.LogRecord('foobar', 20, 'foobar', None, None, None, None)
IGNORED_LOG_ATTRIBUTES = [a for a in _dummy_log_record.__dict__.keys()] + ['message']


class GELFHandler(logging.Handler):
    def __init__(self, timeout=300, *args, **kwargs):
        super(GELFHandler, self).__init__(*args, **kwargs)

        self.request_timeout = timeout

    @property
    def url(self):
        return os.getenv("GELF_URL", None)

    @property
    def hostname(self):
        try:
            return socket.gethostname()
        except Exception:
            return "unknown"

    def emit(self, record, **kwargs):
        if self.url is None:
            logger.error(
                "GELF_URL environment variable needs to be configured to log with "
                "castle.cms.gelf.GELFHandler")
            return

        try:
            short_message = self.format(record)

            full_message = {}
            for name in record.__dict__.keys():
                # ignored attributes are, basically, default attributes of a standard LogRecord
                # as well as the 'message'
                if name not in IGNORED_LOG_ATTRIBUTES:
                    full_message[name] = record.__dict__[name]

            timestamp = int(time.time())
            if 'timestamp' in full_message:
                try:
                    timestamp = int(full_message['timestamp'])
                    del full_message['timestamp']
                except Exception:
                    logger.error(
                        "could get valid timestamp from full_message for GELFHandler",
                        exc_info=True)

            msg = {
                "version": "1.1",
                "host": self.hostname,
                "full_message": full_message,
                "short_message": short_message,
                "level": record.levelno,
                "timestamp": timestamp,
            }

            resp = requests.post(
                self.url,
                data=json.dumps(msg),
                headers={"Content-Type": "application/json"},
                timeout=self.request_timeout)
            if resp.status_code < 200 or resp.status_code >= 300:
                logger.error("GELF endpoint rejecting log: (HTTP {}) {}\n\nMessage:\n{}".format(
                    resp.status_code,
                    resp.content,
                    self.format(record)))
        except Exception:
            logger.error("failed to process log message to GELF url.\n\nMessage:\n{}".format(
                self.format(record)), exc_info=True)



from wildcard.hps.interfaces import IWildcardHPSLayer
from plone.app.contenttypes.interfaces import IPloneAppContenttypesLayer
from plone.app.mosaic.interfaces import IMosaicLayer
from plone.app.tiles.interfaces import ITilesFormLayer
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


class ICastleLayer(IDefaultBrowserLayer, IMosaicLayer, ITilesFormLayer,
                   IPloneAppContenttypesLayer, IWildcardHPSLayer):
    pass


class IVersionViewLayer(ICastleLayer):
    """
    when viewing a version of an object
    """



from plone.namedfile.interfaces import INamedImage
from zope.interface import Attribute
from OFS.interfaces import IApplication
from plone.app.contenttypes.interfaces import IFile
from plone.supermodel import model
from Products.CMFPlone.interfaces import IHideFromBreadcrumbs
from zope.interface import Interface
from zope import schema
from plone.api.portal import get_registry_record
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory


class ICastleApplication(IApplication):
    pass


class IDashboard(IHideFromBreadcrumbs):
    pass


class IMedia(model.Schema, IFile):
    pass


class IVideo(IMedia):
    pass


class IAudio(IMedia):
    pass


@provider(IContextAwareDefaultFactory)
def get_default_text(context):
    return get_registry_record('castle.resource_slide_view_more_link_text', None)


@provider(IContextAwareDefaultFactory)
def get_default_url(context):
    return get_registry_record('castle.resource_slide_view_more_link_url', None)


class ISlideshow(Interface):

    model.fieldset(
        'Resource Slide',
        fields=['show_view_more_link',
                'view_more_link_text',
                'update_default_link_text',
                'view_more_link_url',
                'update_default_link_url']
    )
    custom_dom_id = schema.Text(
        title=u"ID for the slideshow element",
        description=u"If custom styling desired for this slideshow",
        required=False
    )

    show_view_more_link = schema.Bool(
        title=u'Show "View More" link?',
        description=u'The fields below only matter if this box is checked',
        required=True,
        default=False)

    view_more_link_text = schema.TextLine(
        title=u"View More link text",
        description=u'The text a user sees for the optional link at the bottom of the slideshow resource slide. ' # noqa
                    u'This link will be omitted if "View More Link" is unchecked or if the link text or link URL is empty.', # noqa
        required=False,
        defaultFactory=get_default_text)

    update_default_link_text = schema.Bool(
        title=u"Make this the default link text for all slideshows?",
        description=u'This only changes the default value. It will not modify any existing content',
        required=True,
        default=False)

    # directives.widget('make_default_text', SingleCheckBoxFieldWidget)
    view_more_link_url = schema.URI(
        title=u"View More link URL",
        description=u'The URL to which the user is directed for the optional link at the bottom of the slideshow resource slide. ' # noqa
                    u'This link will be omitted if "View More Link" is unchecked or if the link text or link URL is empty.', # noqa
        required=False,
        defaultFactory=get_default_url)

    # directives.widget('make_default_text', SingleCheckBoxFieldWidget)
    update_default_link_url = schema.Bool(
        title=u"Make this the default link url for all slideshows?",
        description=u'This only changes the default value. It will not modify any existing content',
        required=True,
        default=False)


class IParallax(Interface):
    pass


class ITrashed(Interface):
    """
    marker for object that is in the trash
    """


class IHasDefaultImage(Interface):
    """
    marker interface for content to specify that it has default image
    when a manual image is not specified
    """


class IReferenceNamedImage(INamedImage):
    reference = Attribute('')


class IUploadedToYoutube(Interface):
    """Marker interface for videos that have been uploaded to YouTube"""


class ITemplate(Interface):
    """
    marker for object that has been made into template
    """



import logging
import os
from time import time

from AccessControl import getSecurityManager
from Acquisition import aq_parent
from castle.cms import cache
from castle.cms.constants import CASTLE_VERSION_STRING
from castle.cms.events import AppInitializedEvent
from castle.cms.interfaces import ICastleApplication
from celery.result import AsyncResult
from OFS.CopySupport import CopyError
from OFS.CopySupport import _cb_decode
from OFS.CopySupport import _cb_encode
from OFS.CopySupport import eInvalid
from OFS.CopySupport import eNoData
from plone import api
from plone.app.widgets.base import dict_merge
from plone.app.widgets.utils import get_tinymce_options
from plone.app.z3cform.widget import RichTextWidget
from plone.keyring.interfaces import IKeyManager
from plone.session import tktauth
from plone.transformchain.interfaces import ITransform
from Products.CMFPlone.CatalogTool import CatalogTool
from Products.CMFPlone.resources import add_resource_on_request
from ZODB.POSException import ConnectionStateError
from zope.component import getGlobalSiteManager
from zope.component import queryUtility
from zope.component.interfaces import ComponentLookupError
from zope.event import notify
from zope.interface import implementer


logger = logging.getLogger('castle.cms')


def HideSiteLayoutFields_update(self):
    """
    we don't want to hide these fields
    """
    return


def Content_addCreator(self, creator=None):
    """ Add creator to Dublin Core creators.
    """
    if len(self.creators) > 0:
        # do not add creator if one is already set
        return

    if creator is None:
        user = getSecurityManager().getUser()
        creator = user and user.getId()

    # call self.listCreators() to make sure self.creators exists
    if creator and creator not in self.listCreators():
        self.creators = self.creators + (creator, )


def manage_pasteObjects(self, cb_copy_data=None, REQUEST=None):
    """
    override normal paste action to see if we're looking at the paste
    data being stored in cache
    """
    if cb_copy_data is not None:
        cp = cb_copy_data
    elif REQUEST is not None and '__cp' in REQUEST:
        cp = REQUEST['__cp']
    else:
        cp = None
    if cp is None:
        raise CopyError(eNoData)

    try:
        op, mdatas = _cb_decode(cp)
    except Exception:
        raise CopyError(eInvalid)

    try:
        if mdatas[0][0].startswith('cache:'):
            cache_key = mdatas[0][0].replace('cache:', '')
            new_mdatas = cache.get(cache_key)
            cdata = _cb_encode((op, new_mdatas))
            return self._old_manage_pasteObjects(cdata)
    except IndexError:
        pass
    return self._old_manage_pasteObjects(cb_copy_data, REQUEST)


@implementer(ITransform)
class NoopTransform(object):

    order = 8800

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformString(self, result, encoding):
        return None
    transformUnicode = transformIterable = transformString


def AppInitializer_initialize(self):
    self._old_initialize()
    app = self.app[0]
    notify(AppInitializedEvent(app, self.commit))


def SessionPlugin_validateTicket(self, ticket, now=None):
    if now is None:
        now = time()
    if self._shared_secret is not None:
        ticket_data = tktauth.validateTicket(
            self._shared_secret,
            ticket,
            timeout=self.timeout,
            now=now,
            mod_auth_tkt=self.mod_auth_tkt
        )
    else:
        ticket_data = None
        parent = aq_parent(aq_parent(self))

        is_root = ICastleApplication.providedBy(parent)
        if is_root:
            manager = getGlobalSiteManager().queryUtility(IKeyManager)
        else:
            manager = queryUtility(IKeyManager)

        if manager is None:
            return None

        try:
            for secret in manager[u"_system"]:
                if secret is None:
                    continue
                ticket_data = tktauth.validateTicket(
                    secret,
                    ticket,
                    timeout=self.timeout,
                    now=now,
                    mod_auth_tkt=self.mod_auth_tkt
                )
                if ticket_data is not None:
                    break
        except ConnectionStateError:
            logger.warning(
                'Connection state error, swallowing', exc_info=True)
    return ticket_data


def version_overview(self):
    return [CASTLE_VERSION_STRING] + self._old_version_overview()


def rich_text_args(self):
    args = super(RichTextWidget, self)._base_args()
    args['name'] = self.name
    value = self.value and self.value.raw_encoded or ''
    args['value'] = (self.request.get(
        self.field.getName(), value)).decode('utf-8')
    args.setdefault('pattern_options', {})

    # displays tinymce editor for anonymous users
    if api.user.is_anonymous():
        add_resource_on_request(self.request, 'mockup-patterns-tinymce-logged-out')

    merged_options = dict_merge(get_tinymce_options(self.context,
                                                    self.field,
                                                    self.request),  # noqa
                                args['pattern_options'])
    args['pattern_options'] = merged_options

    return args


# AsyncResult objects have a memory leak in them in Celery 4.2.1.
# See https://github.com/celery/celery/pull/4839/
if getattr(AsyncResult, '__del__', False):
    delattr(AsyncResult, '__del__')



from opensearchpy.exceptions import NotFoundError
from wildcard.hps.interfaces import IMappingProvider
from zope.component import getMultiAdapter
from zope.globalrequest import getRequest


CRAWLER_MAPPING = {
    'domain': {
        'type': 'keyword',
        'index': True,
        'store': False
    },
    'sitemap': {
        'type': 'text',
        'index': True,
        'store': False
    },
    'url': {
        'type': 'text',
        'index': False,
        'store': False
    },
    'image_url': {
        'type': 'text',
        'index': False,
        'store': False
    }
}


def index_name(site_index_name):
    return '{site_index_name}_crawler'.format(site_index_name=site_index_name)


def ensure_index_exists(hps, index_name):
    """
    A) make sure index exists
    B) make sure index mapping includes mapped types needed by crawler

    NOTE: if the mapping changes, a new index will need to be generated, and
    then the reindex API operation will need to be used to reindex the
    data, or you'll have to do a full drop of the existing index and then
    initiate a reindex from code and not through the opensearch reindex API
    operation.
    """
    if hps.connection.indices.exists(index_name):
        return

    # create the index
    hps.connection.indices.create(index_name)

    # update the mapping for the crawler
    adapter = getMultiAdapter((getRequest(), hps), IMappingProvider)
    mapping = adapter()
    mapping['properties'].update(CRAWLER_MAPPING)
    hps.connection.indices.put_mapping(body=mapping, index=index_name)


def url_is_indexed(hps, index_name, url):
    try:
        hps.connection.get(index=index_name, id=url)
        return True
    except NotFoundError:
        return False


def index_doc(hps, index_name, url, data):
    hps.connection.index(index=index_name, id=url, body=data)


def delete_from_index(hps, index_name, url):
    try:
        hps.connection.delete(index=index_name, id=url)
    except NotFoundError:
        pass


def get_all_ids(hps, index_name, domain):
    body = {
        "_source": False,  # don't need any data for documents
        "query": {
            "bool": {
                "filter": {
                    "term": {}
                }
            }
        }
    }
    if domain == "archives":
        body["query"]["bool"]["filter"]["term"]["sitemap"] = "archives"
    else:
        body["query"]["bool"]["filter"]["term"]["domain"] = domain

    _ids = []
    scroll_time = '30s'
    result = hps.connection.search(
        index=index_name,
        scroll=scroll_time,
        size=10000,  # hard limit for opensearch
        body=body)
    _ids.extend([r['_id'] for r in result['hits']['hits']])

    # if there aren't 10k records, at leat, no need to scroll
    if len(result['hits']['hits']) < 10000:
        return _ids

    scroll_id = result['_scroll_id']
    while scroll_id:
        result = hps.connection.scroll(scroll_id=scroll_id, scroll=scroll_time)
        if len(result['hits']['hits']) == 0:
            break
        _ids.extend([r['_id'] for r in result['hits']['hits']])
        scroll_id = result['_scroll_id']

    return _ids





"""
Environment Variables:

# see castle.cms.audit
CASTLE_CMS_AUDIT_LOG_INSTANCE

# we assume all logs are submitted to the same index per-instance
# we assume that specific site audit logs can be fetched by including
#   'schema_type', 'schema_version', 'instance', and 'site' as attributes
#   to the search query to this index
CASTLE_CMS_AUDIT_LOG_INDEXNAME

# if audit logs records are coming from a opensearch index that was populated
# by, eg, castle.cms.gelf.GELFHandler, then the log data will be key's on
# an attribute like 'full_message' in the _source results
CASTLE_CMS_AUDIT_LOG_FIELD_MAP_PREFIX

"""
import csv
from cStringIO import StringIO
import logging
import os

from plone import api
from plone.app.uuid.utils import uuidToObject
from Products.CMFPlone.resources import add_resource_on_request
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from castle.cms import audit
from castle.cms.indexing.hps import gen_audit_query
from castle.cms.indexing.hps import health_is_good
from castle.cms.indexing.hps import hps_get_data
from castle.cms.indexing.hps import is_enabled as hps_is_enabled


logger = logging.getLogger("Plone")


class AuditView(BrowserView):
    label = 'My Activity Log'
    limit = 50
    user = True

    inner_template = ViewPageTemplateFile('templates/audit-inner.pt')
    error_template = ViewPageTemplateFile('templates/audit-error.pt')

    def __call__(self):
        # utility function to add resource to rendered page
        add_resource_on_request(self.request, 'castle-components-audit')

        self._user_cache = {}
        self.site_path = '/'.join(self.context.getPhysicalPath())
        try:
            self.results, self.total, _ = self.do_query()

            if 'Export' in self.request.form.get('export', ''):
                return self.export(data=self.results)
        except Exception:
            logger.error("problem with querying hps", exc_info=True)
            self.inner_template = self.error_template
        return super(AuditView, self).__call__()

    def render_inner(self):
        return self.inner_template()

    def get_user(self, userid):
        if not userid:
            return
        if userid in self._user_cache:
            return self._user_cache[userid]
        user = api.user.get(userid)
        if user is not None:
            self._user_cache[userid] = user
        return user

    @property
    def field_map_prefix(self):
        return os.getenv("CASTLE_CMS_AUDIT_LOG_FIELD_MAP_PREFIX", None)

    @property
    def can_connect_to_index(self):
        if not hps_is_enabled():
            return False

        return health_is_good()

    def get_obj(self, uid):
        return uuidToObject(uid)

    def get_path(self, data):
        path = data.get('path')
        if not path:
            return ''
        if path.startswith(self.site_path):
            return path[len(self.site_path):]
        return path

    def get_query(self):
        form = self.request.form

        if self.user:
            user = api.user.get_current().getId().lower()
        else:
            user = form.get('user', '').lower()

        if len(user.strip()) <= 0:
            user = None

        try:
            site_path = '/'.join(api.portal.get().getPhysicalPath())
        except api.exc.CannotGetPortalError:
            site_path = '(zoperoot)'

        query = gen_audit_query(
            field_map_prefix=self.field_map_prefix,
            sort="date",
            sortdir="desc",
            instance=os.getenv("CASTLE_CMS_AUDIT_LOG_INSTANCE"),
            site=site_path,
            typeval=form.get("type", None),
            user=user,
            content=form.get('content', None),
            after=form.get('after', None),
            before=form.get('before', None))

        return query

    def export(self, data=None):
        if data is None:
            index_name = audit.get_index_name()
            query = self.get_query()
            data, _, _ = hps_get_data(index_name, query, size=3000)

        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(['Action', 'Path', 'User', 'Summary', 'Date'])
        for rowdata in data:
            writer.writerow([
                rowdata["actionname"],
                self.get_path(rowdata),
                rowdata["user"],
                rowdata["summary"],
                rowdata["date"]
            ])
        output.seek(0)

        resp = self.request.response
        resp.setHeader('Content-Disposition', 'attachment; filename=export.csv')
        resp.setHeader('Content-Type', 'text/csv')
        return output.read()

    def do_query(self):
        index_name = audit.get_index_name()
        query = self.get_query()

        try:
            page = int(self.request.get('page', 1))
        except Exception:
            page = 1
        start = (page - 1) * self.limit

        results = hps_get_data(index_name, query, from_=start, size=self.limit)

        # process results into something a little less complex to handle within the template
        finalresults = []
        for result in results[0]:
            data = result['_source']
            # IE if the data is coming from a index that had it's content generated by
            # castle.cms.gelf.GELFHandler, then the stored object would actually have
            # a full_message field, which then has the various audit log fields
            if self.field_map_prefix is not None:
                data = data[self.field_map_prefix]
            finalresults.append(data)

        return finalresults, results[1], results[2]



import json

from castle.cms.behaviors.location import ILocation
from castle.cms.behaviors.search import ISearch
from castle.cms.interfaces import ILDData
from castle.cms.utils import is_backend
from castle.cms.utils import site_has_icon
from plone.app.layout.globals.interfaces import IViewView
from plone.tiles import Tile
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFPlone.defaultpage import get_default_page
from Products.CMFPlone.log import logger
from unidecode import unidecode
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.viewlet.interfaces import IViewlet
from zope.viewlet.interfaces import IViewletManager
from castle.cms.constants import CASTLE_VERSION_STRING


head_viewlets = {
    'plone.htmlhead.title': 'plone.htmlhead',
    'plone.links.author': 'plone.htmlhead.links',
    'plone.links.RSS': 'plone.htmlhead.links',
    'plone.htmlhead.dublincore': 'plone.htmlhead',
    'plone.htmlhead.socialtags': 'plone.htmlhead'
}


def _date(obj, date):
    try:
        return getattr(obj, date)().ISO8601()
    except Exception:
        return ''


class MetaDataTile(Tile):

    def get_navigation_links(self):
        return u"""
<link rel="home"
      title="Front page"
      href="{url}"/>
<link rel="contents"
      href="{url}/sitemap"/>""".format(url=self.root_url)

    def get_icons(self):
        if not site_has_icon():
            return ''
        return '''
<link rel="apple-touch-icon" sizes="180x180" href="{url}/site-icon.png">
<link rel="icon" type="image/png" href="{url}/site-icon.png?scale=32" sizes="32x32">
<link rel="icon" type="image/png" href="{url}/site-icon.png?scale=16" sizes="16x16">
<link rel="manifest" href="{url}/manifest.json" crossorigin="use-credentials">
<meta name="theme-color" content="#ffffff">'''.format(url=self.root_url)

    def _wrap_ld(self, data):
        return '<script type="application/ld+json">' + json.dumps(data) + '</script>'

    def _get_robot_config(self, search):
        if(is_backend(self.request)):
            return search.backend_robot_configuration or []
        return search.robot_configuration or []

    def get_ld_data(self):
        result = ''
        ld = ILDData(self.context, None)
        if ld is None:
            return ''
        result += self._wrap_ld(ld.get_data())

        if ISiteRoot.providedBy(self.context):
            try:
                page = self.context[get_default_page(self.context)]
                result += self._wrap_ld(ILDData(page).get_data())
            except AttributeError:
                pass

        return result

    def get_basic_tags(self):
        try:
            context = self.context
            if ISiteRoot.providedBy(context):
                try:
                    context = context[get_default_page(context)]
                except AttributeError:
                    pass
            tags = {
                'modificationDate': _date(context, 'modified'),
                'publicationDate': _date(context, 'effective'),
                'expirationDate': _date(context, 'expires'),
                'generator': CASTLE_VERSION_STRING,
                "distribution": "Global",
            }
            ldata = ILocation(context, None)
            if ldata is not None:
                if ldata.locations:
                    location = ldata.locations
                    if type(location) in (list, tuple, set):
                        location = location[0]
                    tags['location'] = location

            search = ISearch(context, None)
            if search is not None:
                robot_configuration = self._get_robot_config(search)
                config = robot_configuration[:]
                if 'index' not in config:
                    config.append('noindex')
                if 'follow' not in config:
                    config.append('nofollow')
                tags['robots'] = ','.join(config)

            return ''.join([u'<meta name="{}" content="{}">'.format(name, value)
                            for name, value in tags.items()])
        except Exception:
            return u''

    def get_search_link(self):
        return '''
<link rel="search"
      title="Search this site"
      href="{url}/@@search" />'''.format(
            url=self.root_url
          )

    def get_printcss_link(self):
        template = ''' <link rel="stylesheet" href="{url}/++plone++castle/less/public/print.css" type="text/css" media="print">'''  # noqa:E501
        return template.format(url=self.root_url)

    def get_canonical_url(self):
        context_state = getMultiAdapter(
            (self.context, self.request), name=u'plone_context_state')
        canonical_object = context_state.canonical_object()
        if self.context != canonical_object:
            can_state = getMultiAdapter(
                (canonical_object, self.request), name=u'plone_context_state')
            url = can_state.view_url()
        else:
            url = context_state.view_url()
        return u'    <link rel="canonical" href="%s" />' % url

    def __call__(self):
        portal_state = getMultiAdapter((self.context, self.request),
                                       name=u'plone_portal_state')
        self.root_url = portal_state.navigation_root_url()

        result = self.get_canonical_url()
        alsoProvides(self, IViewView)
        for name, manager_name in head_viewlets.items():
            manager = queryMultiAdapter(
                (self.context, self.request, self),
                IViewletManager, name=manager_name
            )
            viewlet = queryMultiAdapter(
                (self.context, self.request, self, manager),
                IViewlet, name=name
            )
            if viewlet is not None:
                try:
                    viewlet.update()
                    result += viewlet.render()
                except Exception:
                    logger.warn('Error rendering head viewlet %s' % name, exc_info=True)
        result += unidecode(self.get_basic_tags())
        result += unidecode(self.get_navigation_links())
        result += unidecode(self.get_ld_data())
        result += unidecode(self.get_icons())
        result += unidecode(self.get_search_link())
        result += unidecode(self.get_printcss_link())
        return u'<html><head>%s</head></html>' % result



from castle.cms.tiles.base import BaseTile
from castle.cms.widgets import ImageRelatedItemFieldWidget
from plone.autoform import directives as form
from plone.supermodel import model
from zope import schema
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class ParallaxTile(BaseTile):

    @property
    def relatedItems(self):
        return self.context.relatedItems

    def get_image(self, is_static):
        if is_static:
            image = self.data.get('static_image')
            if not image:
                return
            return self.utils.get_object(self.data['static_image'][0])
        else:
            image = self.data.get('image')
            if not image:
                return
            return self.utils.get_object(self.data['image'][0])

    def get_horizontal_align(self, margin, is_static):
        if is_static:
            position = self.data.get('static_hor_text_position')
        else:
            position = self.data.get('hor_text_position')
        if position == 'start':
            return '5%' if margin == 'left' else '45%'
        elif position == 'end':
            return '45%' if margin == 'left' else '5%'
        else:
            return '10%'


class IParallaxTileSchema(model.Schema):

    # --------------------- Parallax Tile ---------------------
    model.fieldset(
        'parallax_tile',
        label=u'Parallax Tile',
        fields=[
            'image',
            'bg_color',
            'title',
            'text',
            'text_color',
            'text_shadow',
            'hor_text_position',
        ]
    )

    form.widget(image=ImageRelatedItemFieldWidget)
    image = schema.List(
        title=u'Parallax Image',
        description=u'Image with scrolling parallax effect.',
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )

    bg_color = schema.Choice(
        title=u'Background Color',
        description=u'If no image selected, tile will be this color',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('black', 'black', u'Black'),
            SimpleTerm('white', 'white', u'White'),
        ]),
        default=u'white'
    )

    title = schema.TextLine(
        title=u'Slide Title',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    text = schema.Text(
        title=u'Slide Text',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    text_color = schema.Choice(
        title=u'Text Color',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('black', 'black', u'Black'),
            SimpleTerm('white', 'white', u'White'),
        ]),
        default=u'black'
    )

    text_shadow = schema.Choice(
        title=u'Text Shadow',
        description=u'Adds a shadow to the text over the image.',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('none', 'None', u'None'),
            SimpleTerm('1px 1px 1px black', 'black', u'Black'),
            SimpleTerm('1px 1px 1px white', 'white', u'White'),
        ]),
        default=u'none'
    )

    hor_text_position = schema.Choice(
        title=u'Slide Text Position (Horizontal)',
        description=u'How the text will be horizontally positioned on the tile.',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('start', 'start', u'Left (Half Width)'),
            SimpleTerm('center', 'center', u'Center (Full Width)'),
            SimpleTerm('end', 'end', u'Right (Half Width)'),
        ]),
        default=u'center'
    )

    # --------------------- Static Tile ---------------------
    model.fieldset(
        'static_tile',
        label=u'Static Tile',
        fields=[
            'static_image',
            'static_bg_color',
            'static_title',
            'static_text',
            'static_text_color',
            'static_text_shadow',
            'static_hor_text_position',
        ]
    )

    form.widget(static_image=ImageRelatedItemFieldWidget)
    static_image = schema.List(
        title=u'Slide Image',
        description=u'Image on static tile to create parallax effect.',
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )

    static_bg_color = schema.Choice(
        title=u'Background Color',
        description=u'If no image selected, tile will be this color',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('black', 'black', u'Black'),
            SimpleTerm('white', 'white', u'White'),
        ]),
        default=u'white'
    )

    static_title = schema.TextLine(
        title=u'Slide Title',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    static_text = schema.Text(
        title=u'Slide Text',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    static_text_color = schema.Choice(
        title=u'Text Color',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('black', 'black', u'Black'),
            SimpleTerm('white', 'white', u'White'),
        ]),
        default=u'black'
    )

    static_text_shadow = schema.Choice(
        title=u'Text Shadow',
        description=u'Adds a shadow to the text over the image.',
        required=True,
        vocabulary=SimpleVocabulary([
            SimpleTerm('none', 'None', u'None'),
            SimpleTerm('1px 1px 1px black', 'black', u'Black'),
            SimpleTerm('1px 1px 1px white', 'white', u'White'),
        ]),
        default=u'none'
    )

    static_hor_text_position = schema.Choice(
        title=u'Slide Text Position (Horizontal)',
        description=u'How the text will be horizontally positioned on the tile.',
        vocabulary=SimpleVocabulary([
            SimpleTerm('start', 'start', u'Left (Half Width)'),
            SimpleTerm('center', 'center', u'Center (Full Width)'),
            SimpleTerm('end', 'end', u'Right (Half Width)'),
        ]),
        default=u'center'
    )



from AccessControl import Unauthorized
from Acquisition import aq_base, aq_parent
from wildcard.hps.interfaces import IReindexActive
from OFS.interfaces import IItem
from plone import api
from plone.app.uuid.utils import uuidToCatalogBrain as get_brain
from plone.app.contenttypes.interfaces import IFile, IImage
from plone.dexterity.interfaces import IDexterityContent
from plone.event.interfaces import IEvent
from plone.indexer.decorator import indexer
from plone.uuid.interfaces import IUUID
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.interfaces._content import IFolderish
from ZODB.POSException import POSKeyError
from zope.globalrequest import getRequest

from castle.cms.behaviors.location import ILocation
from castle.cms.interfaces import IHasDefaultImage
from castle.cms.interfaces import IReferenceNamedImage
from castle.cms.interfaces import ITrashed


@indexer(IItem)
def getRawRelatedItems(obj):
    try:
        result = []
        for relation in obj.relatedItems:
            try:
                to_obj = relation.to_object
                if to_obj:
                    uuid = IUUID(to_obj)
                    if uuid:
                        result.append(uuid)
            except AttributeError:
                pass
        return result
    except AttributeError:
        return []


@indexer(IItem)
def getLocation(obj):
    bdata = ILocation(obj, None)
    if bdata and bdata.location:
        return bdata.location[0]


@indexer(IItem)
def hasImage(obj):
    if IHasDefaultImage.providedBy(obj):
        return True
    return getattr(aq_base(obj), 'image', None) is not None


@indexer(IItem)
def image_info(obj):
    try:
        image = obj.image
        if IReferenceNamedImage.providedBy(image):
            data = {'reference': image.reference}
        else:
            width, height = image.getImageSize()
            data = {
                'width': width,
                'height': height,
                'focal_point': [width / 2, height / 2]
            }

            try:
                data['focal_point'] = image.focal_point
            except AttributeError:
                try:
                    data['focal_point'] = obj._image_focal_point
                except Exception:
                    pass
        return data
    except AttributeError:
        pass


@indexer(IItem)
def getContentTypeContent(obj):
    return 'text/html'


@indexer(IFile)
def getContentTypeFile(obj):
    try:
        return obj.file.original_content_type
    except Exception:
        try:
            return obj.file.contentType
        except Exception:
            pass


@indexer(IImage)
def getContentTypeImage(obj):
    try:
        return obj.image.contentType
    except Exception:
        pass


@indexer(IEvent)
def recurrence(obj):
    try:
        return obj.recurrence
    except AttributeError:
        pass


@indexer(IItem)
def trashed(obj):
    while obj:
        if ITrashed.providedBy(obj):
            return True
        obj = aq_parent(obj)
    return False


@indexer(IDexterityContent)
def last_modified_by(context):
    req = getRequest()
    creator = getattr(aq_base(context), '_last_modified_by', None)
    if req is not None and not IReindexActive.providedBy(req):
        # here, we're assuming the current user editing the object
        # is the logged in user performing the current action
        creator = api.user.get_current().getId()
        # also set on context so we can retrieve later when reindexing
        context._last_modified_by = creator

    if creator is None:
        try:
            creator = context.Creators()[0]
        except Exception:
            creator = 'admin'

    rt = api.portal.get_tool("portal_repository")

    if rt is None or not rt.isVersionable(context):
        # not versionable; fallback to the creator
        return creator

    try:
        history = rt.getHistoryMetadata(context)
    except Unauthorized:
        history = None
    if not history:
        return creator

    try:
        if not rt.isUpToDate(context):
            return creator
    except POSKeyError:
        return creator

    length = history.getLength(countPurged=False)

    last = history.retrieve(length - 1)
    if not last or type(last) != dict:
        # unexpected version metadata; fallback to the creator
        return creator

    metadata = last.get("metadata")
    if not metadata or type(metadata) != dict:
        # unexpected version metadata; fallback to the creator
        return creator

    sys_metadata = metadata.get("sys_metadata")
    if not sys_metadata or type(sys_metadata) != dict:
        # unexpected version metadata; fallback to the creator
        return creator

    principal = sys_metadata.get("principal")
    if not principal or type(principal) != str:
        # unexpected version metadata; fallback to the creator
        return creator

    return principal


@indexer(IItem)
def has_private_parents(obj):
    if (api.content.get_state(obj) != 'published'):
        return True  # needs to be True for private self as well as parents
    parent = aq_parent(obj)
    while not ISiteRoot.providedBy(parent):
        try:
            parent_brain = get_brain(parent.UID())
            try:
                if parent_brain.has_private_parents:
                    return True
            except AttributeError:
                if api.content.get_state(parent) != 'published':
                    return True
        except Exception:
            pass  # to be extra secure, could return True here. Better to be fault tolerant for now.
        parent = aq_parent(parent)
    return False


@indexer(IItem)
def self_or_child_has_title_description_and_image(obj):
    if (IFolderish.providedBy(obj)):
        contents = obj.getFolderContents()
        for item in contents:
            if item.get('self_or_child_has_title_description_and_image', False):
                return True
    return bool(getattr(aq_base(obj), 'title', False)) and \
        bool(getattr(aq_base(obj), 'description', False)) and \
        bool(getattr(aq_base(obj), 'image', False))


@indexer(IImage)
def has_custom_markup(image):
    if image.custom_markup:
        return True



from castle.cms.interfaces import ICrawlerConfiguration
from castle.cms.utils import get_public_url
from castle.cms.indexing import hps

from DateTime import DateTime
from plone import api
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.resources import add_resource_on_request
from Products.Five import BrowserView
from urlparse import urlparse
from zope.component import getUtility
import dateutil.parser

import datetime
import json
import Missing


def custom_json_handler(obj):
    if obj == Missing.Value:
        return None
    if type(obj) in (datetime.datetime, datetime.date):
        return obj.isoformat()
    if type(obj) == DateTime:
        return obj.ISO8601()
    return obj


class Search(BrowserView):
    def __call__(self):
        # utility function to add resource to rendered page
        add_resource_on_request(self.request, 'castle-components-search')
        return super(Search, self).__call__()

    @property
    def options(self):
        search_types = [
            {
                'id': 'images',
                'label': 'Image',
                'query': {
                    'portal_type': 'Image'
                }
            },
            {
                'id': 'page',
                'label': 'Page',
                'query': {
                    'portal_type': ['Document', 'Folder']
                }
            }
        ]

        ptypes = api.portal.get_tool('portal_types')
        allow_anyway = ['Audio']
        for type_id in ptypes.objectIds():
            if type_id in ('Link', 'Document', 'Folder'):
                continue
            _type = ptypes[type_id]
            if not _type.global_allow and type_id not in allow_anyway:
                continue
            search_types.append({
                'id': type_id.lower(),
                'label': _type.title,
                'query': {
                    'portal_type': type_id
                }
            })

        additional_sites = []
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ICrawlerConfiguration, prefix='castle')
        if hps.is_enabled() and settings.crawler_active and settings.crawler_site_maps:
            result = hps.get_index_summary(hps.get_index_name(), dict(field="domain"))
            for res in result:
                site_name = res.get('key')
                if '.' not in site_name or 'amazon' in site_name:
                    continue
                additional_sites.append(site_name)

        parsed = urlparse(get_public_url())
        return json.dumps({
            'searchTypes': sorted(search_types, key=lambda st: st['label']),
            'additionalSites': sorted(additional_sites),
            'currentSiteLabel': parsed.netloc,
            'searchHelpText': api.portal.get_registry_record('castle.search_page_help_text', None),
        })

    @property
    def search_url(self):
        if api.user.is_anonymous():
            try:
                url = api.portal.get_registry_record('castle.searchurl')
                if url:
                    return url
            except Exception:
                pass
        return '%s/@@searchajax' % (
            self.context.absolute_url()
        )


_search_attributes = [
    'Title',
    'Description',
    'Subject',
    'contentType',
    'created',
    'modified',
    'effective',
    'hasImage',
    'is_folderish',
    'portal_type',
    'review_state',
    'path.path'
]

_valid_params = [
    'SearchableText',
    'portal_type',
    'Subject',
    'Subject:list',
    'after',
    'sort_on',
    'sort_order'
]


class SearchAjax(BrowserView):

    def __call__(self):
        self.catalog = api.portal.get_tool('portal_catalog')
        self.request.response.setHeader('Content-type', 'application/json')

        query = {}
        for name in _valid_params:
            real_name = name
            if real_name.endswith(':list'):
                real_name = real_name[:-len(':list')]
            if self.request.form.get(name):
                query[real_name] = self.request.form[name]
            elif self.request.form.get(name + '[]'):
                query[real_name] = self.request.form[name + '[]']

        if query.get('after'):
            if query.get('sort_on') not in ('effective', 'modified', 'created'):
                sort_on = query['sort_on'] = 'effective'
            else:
                sort_on = query['sort_on']
            try:
                date = dateutil.parser.parse(query.pop('after'))
                start = DateTime(date)
                query[sort_on] = {
                    'query': start,
                    'range': 'min'
                }
            except (KeyError, AttributeError, ValueError, TypeError):
                pass

        query['review_state'] = 'published'

        registry = getUtility(IRegistry)
        if not registry.get('plone.allow_public_in_private_container', False):
            query['has_private_parents'] = False
        query['exclude_from_search'] = False

        try:
            page_size = int(self.request.form.get('pageSize'))
        except Exception:
            page_size = 20
        page_size = min(page_size, 50)
        try:
            page = int(self.request.form.get('page'))
        except Exception:
            page = 1

        if hps.is_enabled():
            return self.get_hps_results(page, page_size, query)
        else:
            return self.get_results(page, page_size, query)

    def get_results(self, page, page_size, query):
        # regular plone search
        site_path = '/'.join(self.context.getPhysicalPath())
        start = (page - 1) * page_size
        end = start + page_size
        catalog = api.portal.get_tool('portal_catalog')
        raw_results = catalog(**query)
        items = []

        registry = getUtility(IRegistry)
        view_types = registry.get('plone.types_use_view_action_in_listings', [])

        for brain in raw_results[start:end]:
            attrs = {}
            for key in _search_attributes:
                attrs[key] = getattr(brain, key, None)
            url = base_url = brain.getURL()
            if brain.portal_type in view_types:
                url += '/view'
            attrs.update({
                'path': brain.getPath()[len(site_path):],
                'base_url': base_url,
                'url': url
            })
            items.append(attrs)

        return json.dumps({
            'count': len(raw_results),
            'results': items,
            'page': page,
            'suggestions': []
        }, default=custom_json_handler)

    def get_hps_results(self, page, page_size, query):
        # returns json:
        #   - count -- total hits
        #   - results -- list of items:
        #       - review_state -- always 'published'
        #       - score
        #       - path
        #       - base_url
        #       - url
        #   - page
        #   - suggestions -- list of objects from opensearch

        results = hps.get_search_results(
            self.context,
            self.request,
            self.catalog,
            _search_attributes,
            page,
            page_size,
            query)

        return json.dumps(results, default=custom_json_handler)




from collections import namedtuple

from castle.cms.package import CASTLE_VERSION  # noqa:F401
from castle.cms.package import CASTLE_VERSION_STRING  # noqa:F401
from re import compile

SHIELD = namedtuple('SHIELD', 'NONE BACKEND ALL')('', 'backend', 'all')
MAX_PASTE_ITEMS = 40
ALL_SUBSCRIBERS = '--SUBSCRIBERS--'
ALL_USERS = '--USERS--'
DEFAULT_SITE_LAYOUT_REGISTRY_KEY = 'castle.cms.mosaic.default_site_layout'
CRAWLED_DATA_KEY = 'castle.cms.crawldata'
TRASH_LOG_KEY = 'castle.cms.empty-trash-log'
ANONYMOUS_USER = "Anonymous User"
DEFAULT_FONT_SIZE_SMALL = u'12pt'
DEFAULT_FONT_SIZE_MEDIUM = u'14pt'
DEFAULT_FONT_SIZE_LARGE = u'16pt'
ABSOLUTE_FONT_UNITS = (
    'cm',
    'mm',
    'Q',
    'in',
    'pc',
    'pt',
    'px',
)
VALID_CSS_FONT_SIZE_PATTERN = compile("^\s*\d+\s*({})\s*$".format(  # noqa:W605
    '|'.join(ABSOLUTE_FONT_UNITS)
))



import logging

from castle.cms.behaviors.search import ISearch
from castle.cms.social import COUNT_ANNOTATION_KEY

from wildcard.hps import mapping
from wildcard.hps import query
from wildcard.hps.interfaces import IAdditionalIndexDataProvider
from plone import api
from zope.annotation.interfaces import IAnnotations
from zope.interface import implements


logger = logging.getLogger("Plone")


class MappingAdapter(mapping.MappingAdapter):
    _default_mapping = mapping.MappingAdapter._default_mapping.copy()
    _default_mapping.update({
        'page_views': {'store': True, 'type': 'integer', 'null_value': 0},
        'facebook_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'twitter_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'linkedin_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'pinterest_shares': {'store': True, 'type': 'integer',
                             'null_value': 0},
        'searchterm_pins': {'store': True, 'type': 'text',
                            'index': False},
        'contributors': {'store': False, 'type': 'text',
                         'index': True},
        'immediate_folder': {'store': True, 'type': 'text',
                             'index': False},
        'parent_folder': {'store': True, 'type': 'keyword',
                          'index': False}
    })


class AdditionalIndexDataProvider(object):
    implements(IAdditionalIndexDataProvider)

    def __init__(self, obj):
        self.obj = obj

    def __call__(self, hpscatalog, existing_data):
        annotations = IAnnotations(self.obj)
        data = {}
        counts = annotations.get(COUNT_ANNOTATION_KEY, {})
        for key, value in counts.items():
            key = key.replace('_matomo', '')
            if isinstance(value, dict):
                value = value.get('total') or 0
            if key in ('page_views',):
                data[key] = value
            else:
                data[key + '_shares'] = value
        sdata = ISearch(self.obj, None)
        if sdata:
            pins = getattr(sdata, 'searchterm_pins', []) or []
            for i, pin in enumerate(pins):
                if not pin:
                    continue
                p = pin.lower()
                if type(p) != unicode:
                    p = unicode(p, 'utf-8')
                pins[i] = p
            data['searchterm_pins'] = pins
        else:
            data['searchterm_pins'] = []

        try:
            existing_searchable_text = existing_data.get('SearchableText', u'')
            if type(existing_searchable_text) != unicode:
                existing_searchable_text = unicode(existing_searchable_text, 'utf-8')
            data['SearchableText'] = u'%s %s' % (existing_searchable_text, u' '.join(data['searchterm_pins']))
        except UnicodeError:
            logger.error("unicode error when generating SearchableText", exc_info=True)

        try:
            data['contributors'] = list(
                self.obj.creators + self.obj.contributors)
        except Exception:
            pass
        path = self.obj.getPhysicalPath()
        data['parent_folder'] = '/'.join(path[:-1])
        site_path = api.portal.get().getPhysicalPath()
        if len(path) > (len(site_path) + 1):
            data['immediate_folder'] = path[len(site_path):][0]
        else:
            data['immediate_folder'] = '/'
        return data


class QueryAssembler(query.QueryAssembler):
    def __call__(self, dquery):
        dquery['trashed'] = False
        query = super(QueryAssembler, self).__call__(dquery)
        # take into account views, likes and custom weighting
        try:
            searchq = dquery.get('SearchableText', '')
            if isinstance(searchq, dict):
                searchq = searchq.get('query', '')
            searchq = searchq.lower().strip('*')
            query = {
                'script_score': {
                    'query': query,
                    # "boost_mode": "sum",  # add score and modified score,
                    'script': {
                        'lang': 'painless',
                        'source': '''int max_shares = 5000;
                                        int max_popularity = 200000;
                                        String[] socialFields = new String[4];
                                        socialFields[0] = 'twitter';
                                        socialFields[1] = 'facebook';
                                        socialFields[2] = 'pinterest';
                                        socialFields[3] = 'linkedin';

                                        float boost = 1.0f;
                                        float max_boost = 2.5f;
                                        long shareCount = 0;

                                        for (int i=0; i<socialFields.length; i++) {
                                        String key = socialFields[i] + '_shares';
                                        if(doc[key].size() != 0){
                                            long docValue = doc[key].value;
                                            shareCount += docValue;
                                        }
                                        }

                                        boost += (shareCount / max_shares);
                                        if (doc['page_views'].size() != 0) {
                                        long docValue = doc['page_views'].value;
                                        boost += (docValue / max_popularity);
                                        }
                                        boost = (float)Math.min(boost, max_boost);
                                        return boost;'''
                    }
                }
            }
        except KeyError:
            pass
        return query



from castle.cms.tiles.base import BaseTile
from castle.cms.widgets import ImageRelatedItemFieldWidget
from castle.cms.widgets import SlideRelatedItemsFieldWidget
from castle.cms.widgets import VideoRelatedItemsFieldWidget
from plone.autoform import directives as form
from plone.supermodel import model
from zope import schema
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class SlideTile(BaseTile):

    @property
    def relatedItems(self):
        return self.context.relatedItems

    @property
    def slide_type(self):
        return self.data.get('display_type', 'background-image')


class ISlideTileSchema(model.Schema):

    model.fieldset(
        'type_and_text',
        label=u'Type & Text',
        fields=[
            'display_type',
            'title',
            'text',
        ]
    )

    display_type = schema.Choice(
        title=u"Display Type",
        vocabulary=SimpleVocabulary([
            SimpleTerm('background-image', 'background-image', u'Background Image'),  # noqa
            SimpleTerm('left-image-right-text','left-image-right-text', u'Left Image Right Text'),  # noqa
            SimpleTerm('background-video', 'background-video', u'Background Video'),  # noqa
            SimpleTerm('left-video-right-text','left-video-right-text', u'Left Video Right Text'),  # noqa
            SimpleTerm('resource-slide','resource-slide', u'Resource Slide')  # noqa
        ]),
        required=True,
        default='background-image'
    )

    title = schema.TextLine(
        title=u'Slide Title',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    text = schema.Text(
        title=u'Slide Text',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    #######################################################################################################
    model.fieldset(
        'text_positioning',
        label=u'Text Positioning',
        fields=[
            'hor_text_position',
            'vert_text_position',
            'text_alignment',
            'justify_wrapped_text',
        ]
    )

    hor_text_position = schema.Choice(
        title=u"Slide Text Position (Horizontal)",
        description=u'This setting only applies to large screens. '
                    u'On small screens, Center (Full Width) will always be deisplayed.',
        vocabulary=SimpleVocabulary([
            SimpleTerm('start', 'start', u'Left (Half Width)'),
            SimpleTerm('center', 'center', u'Center (Full Width)'),
            SimpleTerm('end', 'end', u'Right (Half Width)'),
        ]),
        required=True,
        default=u'center'
    )

    vert_text_position = schema.Choice(
        title=u"Slide Text Position (Vertical)",
        vocabulary=SimpleVocabulary([
            SimpleTerm('top', 'top', u'Top'),
            SimpleTerm('middle', 'middle', u'Middle'),
            SimpleTerm('bottom', 'bottom', u'Bottom'),
        ]),
        required=True,
        default=u'middle'
    )

    text_alignment = schema.Choice(
        title=u"Slide Text Alignment",
        description=u'The alignment of slide text relative to other text on the page. '
                    u'This does not change the position of the text section selected above. '
                    u'(Note: this is ignored on smaller screens)',
        vocabulary=SimpleVocabulary([
            SimpleTerm('left', 'left', u'Left'),
            SimpleTerm('center', 'center', u'Center'),
        ]),
        required=True,
        default=u'center'
    )

    justify_wrapped_text = schema.Bool(
        title=u'Justify Wrapped Text Lines',
        description=u'Select this option to force any text line that wraps to more than one line to take up the entire width. ' # noqa
                    u'This will assure that there are no gaps to the left and/or right when text fills the whole line, regardless of Text Alignment Choice above.', # noqa 
        default=False,
    )

    #######################################################################################################
    model.fieldset(
        'media_settings',
        label=u'Media Settings',
        fields=[
            'image',
            'video',
        ]
    )

    form.widget(image=ImageRelatedItemFieldWidget)
    image = schema.List(
        title=u"Slide Image",
        description=u"Reference image on the site.",
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )

    form.widget(video=VideoRelatedItemsFieldWidget)
    video = schema.List(
        title=u"Slide Video",
        description=u"Reference video on the site.",
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )

    #######################################################################################################
    model.fieldset(
        'related_items',
        label=u'Related Items',
        fields=['related_items']
    )

    form.widget('related_items', SlideRelatedItemsFieldWidget)
    related_items = schema.List(
        title=u"Related Items",
        description=u"Items to include on Resource Slide. "
                    u"To be selectable, a content item must contain a Title, Description (Summary), and a Lead Image.", # noqa
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )
